"""Stopping criteria for nested sampling.

This module defines various stopping criteria for nested sampling algorithms.
"""

import operator
from dataclasses import dataclass
from typing import Dict, Literal, Union

COMPARISON_OPERATORS = {
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
    "==": operator.eq,
    "!=": operator.ne,
}


class StoppingCriterionRegistry:
    """
    Registry for stopping criteria.

    This class is used to register and retrieve stopping criteria classes.
    It allows for easy extension and management of different stopping criteria.
    """

    _registry = {}

    @classmethod
    def register(cls, *names: str):
        def decorator(criterion_cls):
            for name in names:
                cls._registry[name.lower()] = criterion_cls
            return criterion_cls

        return decorator

    @classmethod
    def get(cls, name: str, **kwargs):
        key = name.lower()
        if key not in cls._registry:
            raise ValueError(f"No registered criterion with name '{name}'")
        return cls._registry[key](**kwargs)

    @classmethod
    def list_available(cls):
        return list(cls._registry.keys())


@dataclass
class StoppingCriterion:
    """Class for defining stopping criteria.

    Parameters
    ----------
    name : str
        Name of the stopping criterion.
    tolerance : float
        Tolerance value for the stopping criterion.
    comparison : str
        Comparison operator for the stopping criterion.
            Valid options are: '<', '>', '<=', '>=', '==', '!='.
    """

    name: str
    tolerance: float
    comparison: Literal["<", ">", "<=", ">=", "==", "!="]

    def __post_init__(self):
        if self.comparison not in COMPARISON_OPERATORS:
            raise ValueError(f"Invalid comparison operator: {self.comparison}")
        self._operator = COMPARISON_OPERATORS[self.comparison]

    def is_met(self, value: Union[float, int]) -> bool:
        """
        Check if the stopping criterion is met.

        Parameters
        ----------
        value : float or int
            The value to compare against the stopping criterion.

        Returns
        -------
        bool
            True if the stopping criterion is met, False otherwise.
        """
        return self._operator(value, self.tolerance)

    def __and__(self, other: "CriterionGroup") -> "CriterionGroup":
        return CriterionGroup([self]) & other

    def __or__(self, other: "CriterionGroup") -> "CriterionGroup":
        return CriterionGroup([self]) | other


@dataclass
class CriterionGroup:
    criteria: list
    mode: Literal["and", "or"] = "and"

    def is_met(self, values: Dict[str, float]) -> bool:
        results = [c.is_met(values[c.name]) for c in self.criteria]
        return all(results) if self.mode == "and" else any(results)

    def __and__(
        self, other: Union[StoppingCriterion, "CriterionGroup"]
    ) -> "CriterionGroup":
        return CriterionGroup(
            self.criteria
            + (
                [other]
                if isinstance(other, StoppingCriterion)
                else other.criteria
            ),
            mode="and",
        )

    def __or__(
        self, other: Union[StoppingCriterion, "CriterionGroup"]
    ) -> "CriterionGroup":
        return CriterionGroup(
            self.criteria
            + (
                [other]
                if isinstance(other, StoppingCriterion)
                else other.criteria
            ),
            mode="or",
        )

    @property
    def names(self) -> list[str]:
        """
        Get the names of the stopping criteria in the group.

        Returns
        -------
        list[str]
            List of names of the stopping criteria.
        """
        return [criterion.name for criterion in self.criteria]

    @property
    def tolerances(self) -> Dict[str, float]:
        """
        Get the tolerances of the stopping criteria in the group.

        Returns
        -------
        Dict[str, float]
            Dictionary mapping criterion names to their tolerances.
        """
        return {
            criterion.name: criterion.tolerance for criterion in self.criteria
        }


@StoppingCriterionRegistry.register("ess")
class ESS(StoppingCriterion):
    """
    Effective sample size (ESS) stopping criterion.

    Parameters
    ----------
    tolerance : float
        Tolerance value for the stopping criterion.
    """

    def __init__(self, tolerance: float = 5000.0):
        super().__init__("ess", tolerance, comparison=">=")


@StoppingCriterionRegistry.register("log_evidence_ratio", "ratio")
class LogEvidenceRatio(StoppingCriterion):
    """
    Log ratio of evidence between live points and all samples.

    Parameters
    ----------
    tolerance : float
        Tolerance value for the stopping criterion.
    """

    def __init__(self, tolerance: float = 0.0):
        super().__init__("log_evidence_ratio", tolerance, comparison="<=")


@StoppingCriterionRegistry.register(
    "log_evidence_ratio_nested_samples", "ratio_ns"
)
class LogEvidenceRatioNestedSamples(StoppingCriterion):
    """
    Log ratio of evidence between live point and nested samples.

    Parameters
    ----------
    tolerance : float
        Tolerance value for the stopping criterion.
    """

    def __init__(self, tolerance: float = 0.0):
        super().__init__(
            "log_evidence_ratio_nested_samples", tolerance, comparison="<="
        )


@StoppingCriterionRegistry.register("Z_err", "evidence_error")
class EvidenceError(StoppingCriterion):
    """
    Evidence error stopping criterion.

    Parameters
    ----------
    name : str
        Name of the stopping criterion.
    tolerance : float
        Tolerance value for the stopping criterion.
    comparison : str
        Comparison operator for the stopping criterion.
    """

    def __init__(self, tolerance: float = 0.1):
        super().__init__("evidence_error", tolerance, comparison="<=")


@StoppingCriterionRegistry.register("dlogZ", "difference_log_evidence")
class DifferenceLogEvidence(StoppingCriterion):
    """Difference in log evidence stopping criterion.

    This is the standard nested sampling stopping criterion.

    Parameters
    ----------
    tolerance : float
        Tolerance value for the stopping criterion.
    """

    def __init__(self, tolerance: float = 0.1):
        super().__init__("difference_log_evidence", tolerance, comparison="<=")


@StoppingCriterionRegistry.register("fractional_error")
class FractionalError(StoppingCriterion):
    """
    Fractional error stopping criterion.

    Parameters
    ----------
    tolerance : float
        Tolerance value for the stopping criterion.
    """

    def __init__(self, tolerance: float = 0.1):
        super().__init__("fractional_error", tolerance, comparison="<=")



import os
import subprocess
import sys

import nessai

# -- Run pre-build steps script ----------------------------------------------


def run_external_script(script_path):
    """Run a script to generate plots."""
    script_path = os.path.abspath(script_path)
    subprocess.run(["python", script_path], check=True)


# Generate plots from an example run
run_external_script("scripts/example_run.py")

# -- Path setup --------------------------------------------------------------

sys.path.insert(0, os.path.abspath("../src/nessai/"))

# -- Project information -----------------------------------------------------

project = "nessai"
copyright = "2020, Michael J. Williams"
author = "Michael J. Williams"
version = nessai.__version__
release = nessai.__version__

# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.mathjax",
    "numpydoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    "sphinx.ext.inheritance_diagram",
    "sphinx_tabs.tabs",
    "autoapi.extension",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_book_theme"
html_static_path = []
html_title = "nessai"
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/mj-will/nessai",
    "repository_branch": "main",
    "use_edit_page_button": True,
    "use_issues_button": True,
    "use_repository_button": True,
    "use_download_button": True,
}

# -- Configure autoapi -------------------------------------------------------

autoapi_type = "python"
autoapi_dirs = ["../src/nessai/"]
autoapi_add_toctree_entry = False
autoapi_options = ["members", "show-inheritance", "show-module-summary"]

# -- RST config --------------------------------------------------------------
# Inline python highlighting, base on https://stackoverflow.com/a/68931907

rst_prolog = """
.. role:: python(code)
    :language: python
    :class: highlight
"""



import pytest

from nessai.proposal.flowproposal.base import BaseFlowProposal


def test_conifgure_poolsize(proposal):
    BaseFlowProposal.configure_poolsize(proposal, 10, True, 10)
    assert proposal._poolsize == 10
    assert proposal._poolsize_scale == 1
    assert proposal.update_poolsize is True
    assert proposal.max_poolsize_scale == 10
    assert proposal.ns_acceptance == 1.0


def test_config_poolsize_none(proposal):
    """
    Test the popluation configuration raises an error if poolsize is None.
    """
    with pytest.raises(RuntimeError, match=r"Must specify `poolsize`"):
        BaseFlowProposal.configure_poolsize(
            proposal,
            None,
            True,
            10,
        )


@pytest.mark.parametrize(
    "plot, plot_pool, plot_train",
    [
        (True, True, True),
        ("all", "all", "all"),
        ("train", False, "all"),
        ("pool", "all", False),
        ("min", True, True),
        ("minimal", True, True),
        (False, False, False),
        ("some", False, False),
    ],
)
def test_configure_plotting(proposal, plot, plot_pool, plot_train):
    """Test the configuration of plotting settings"""
    BaseFlowProposal.configure_plotting(proposal, plot)
    assert proposal._plot_pool == plot_pool
    assert proposal._plot_training == plot_train


def test_update_flow_proposal(proposal):
    """Assert the number of inputs is updated"""
    proposal.flow_config = {"model_config": {}}
    proposal.rescaled_dims = 4
    BaseFlowProposal.update_flow_config(proposal)
    assert proposal.flow_config["n_inputs"] == 4


def test_flow_config(proposal):
    """Assert the correct config is returned"""
    config = {"a": 1}
    proposal._flow_config = config
    assert BaseFlowProposal.flow_config.__get__(proposal) is config



# -*- coding: utf-8 -*-
"""Base nested sampler object"""

import datetime
import logging
import os
import pickle
import random
import time
from abc import ABC, abstractmethod
from typing import Any, Callable, Optional, Union
from warnings import warn

import numpy as np
import torch
from glasflow import __version__ as glasflow_version

from .. import __version__ as version
from ..model import Model
from ..utils import safe_file_dump

logger = logging.getLogger(__name__)


class BaseNestedSampler(ABC):
    """Base nested sampler class.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model
    output : str, optional
        Path for the output. If not specified the current working directory
        is used.
    seed : int, optional
        Seed used to seed numpy and torch.
    checkpointing : bool, optional
        Boolean to toggle checkpointing, must be enabled to resume the sampler.
        If false the sampler is still saved at the end of sampling.
    checkpoint_interval : int
        The interval used for checkpointing. By default this is a time interval
        in seconds. If :code:`checkpoint_on_iteration=True` this corresponds to
        the number of iterations between checkpointing.
    checkpoint_on_iteration : bool
        If true the checkpointing interval is checked against the number of
        iterations
    checkpoint_callback : Callback
        Callback function to be used instead of the default function. The
        function will be called in the code:`checkpoint` method as:
        :code`checkpoint_callback(self)`.
    logging_interval : int, optional
        The interval in seconds used for periodic logging. If not specified,
        then periodic logging is disabled.
    log_on_iteration : bool
        If true logging will occur based on the iteration. If false logging
        will be periodic if `logging_interval` is set. In case where neither
        logging is enabled, `log_on_iteration` will be set to true with an
        interval of :code:`nlive`.
    resume_file : str, optional
        Name of the file the sampler will be saved to and resumed from.
    plot : bool, optional
        Boolean to enable or disable plotting.
    n_pool : int, optional
        Number of threads to when for creating the multiprocessing pool.
    pool : object
        User defined multiprocessing pool that will be used when evaluating
        the likelihood.
    """

    def __init__(
        self,
        model: Model,
        nlive: int,
        output: str = None,
        seed: int = None,
        rng: Optional[np.random.Generator] = None,
        checkpointing: bool = True,
        checkpoint_interval: int = 600,
        checkpoint_on_iteration: bool = False,
        checkpoint_callback: Optional[Callable] = None,
        logging_interval: int = None,
        log_on_iteration: bool = True,
        resume_file: str = None,
        plot: bool = True,
        n_pool: Optional[int] = None,
        pool: Optional[Any] = None,
    ):
        logger.info("Initialising nested sampler")

        self.info_enabled = logger.isEnabledFor(logging.INFO)
        self.debug_enabled = logger.isEnabledFor(logging.DEBUG)
        self.configure_rng(seed, rng)
        model.set_rng(self.rng)
        model.verify_model()
        self.n_pool = n_pool
        self.model = model
        self.model.configure_pool(pool=pool, n_pool=n_pool)

        self.nlive = nlive
        self.plot = plot
        self.checkpointing = checkpointing
        self.checkpoint_interval = checkpoint_interval
        self.checkpoint_on_iteration = checkpoint_on_iteration
        self.checkpoint_callback = checkpoint_callback
        if self.checkpoint_on_iteration:
            self._last_checkpoint = 0
        else:
            self._last_checkpoint = datetime.datetime.now()
        self.sampling_time = datetime.timedelta()
        self.sampling_start_time = datetime.datetime.now()
        self.iteration = 0
        self.finalised = False
        self.resumed = False

        self.configure_output(output, resume_file=resume_file)
        self.configure_periodic_logging(logging_interval, log_on_iteration)

        self.live_points = None
        self.history = None

    @property
    def current_sampling_time(self):
        if self.finalised:
            return self.sampling_time
        else:
            return self.sampling_time + (
                datetime.datetime.now() - self.sampling_start_time
            )

    @property
    def likelihood_evaluation_time(self):
        """Current log-likelihood time"""
        return self.model.likelihood_evaluation_time

    @property
    def total_likelihood_evaluations(self):
        """Total number of likelihood evaluations"""
        return self.model.likelihood_evaluations

    likelihood_calls = total_likelihood_evaluations
    """Alias for :code:`total_likelihood_evaluations`"""

    @property
    @abstractmethod
    def posterior_effective_sample_size(self):
        """The effective sample size of the posterior distribution"""
        raise NotImplementedError()

    def configure_output(
        self, output: Union[str, None], resume_file: Union[str, None] = None
    ):
        """Configure the output folder

        Parameters
        ----------
        output : str
            Directory where the results will be stored
        resume_file : str, optional
            Specific file to use for checkpointing. If not specified the
            default is used (nested_sampler_resume.pkl)
        """
        if output is None:
            output = os.getcwd()
        if not os.path.exists(output):
            os.makedirs(output, exist_ok=True)

        if resume_file is None:
            resume_file = os.path.join(output, "nested_sampler_resume.pkl")
        else:
            resume_file = os.path.join(output, resume_file)

        self.output = output
        self.resume_file = resume_file

    def update_output(self, output: str) -> None:
        """Update the output directory and resume file.

        Parameters
        ----------
        output: str
            Path to the output directory
        """
        self.output = output
        resume_file = os.path.split(self.resume_file)[1]
        self.resume_file = os.path.join(output, resume_file)

    def configure_rng(
        self,
        seed: Optional[int] = None,
        rng: Optional[np.random.Generator] = None,
    ):
        """Configure the random number generation.

        If a seed is not specified, a random seed is generated using the
        specified random number generator, or the default numpy random number
        if not specified. In the latter case, the seed is then used to seed
        the new numpy random number generator.

        ..versionadded:: 0.14.0

        Parameters
        ----------
        seed : Optional[int]
            The random seed. If not specified, a random seed is generated.
        rng: Optional[np.random.Generator]
            Random number generator. If not specified, the default numpy
            random number generator is used with the specified seed.
        """
        if seed is None:
            logger.debug("Seed not specified, generating random seed")
            if rng is None:
                seed = random.randint(0, min(np.iinfo(int).max, 2**32 - 1))
            else:
                seed = rng.integers(0, min(np.iinfo(int).max, 2**32 - 1))
        logger.debug(f"Setting random seed to {seed}")
        self.seed = seed
        if rng is None:
            logger.debug(
                f"No rng specified, using the default rng with seed={seed}."
            )
            rng = np.random.default_rng(self.seed)
        self.rng = rng
        torch.manual_seed(self.seed)

    def configure_random_seed(self, seed: Optional[int] = None) -> None:
        """Initialise the random seed

        ..deprecated:: 0.14.0
            Deprecated in favour of
            :py:meth:`~nessai.samplers.base.BaseNestedSampler.configure_rng`.

        Parameters
        ----------
        seed : Optional[int]
            The random seed. If not specified, a random seed is generated.
        """
        msg = (
            "`configure_random_seed` is deprecated and will be removed in a "
            "future release. Use `configure_rng` instead."
        )
        warn(msg, FutureWarning)
        self.configure_rng(seed=seed)

    def configure_periodic_logging(self, logging_interval, log_on_iteration):
        """Configure the periodic logging.

        Parameters
        ----------
        logging_interval : int, optional
            The interval in seconds used for periodic logging. If not
            specified, then periodic logging is disabled.
        log_on_iteration : bool
            If true logging will occur based on the iteration. If false logging
            will be periodic if `logging_interval` is set. In case where
            neither logging is enabled, `log_on_iteration` will be set to true
            with an interval of :code:`nlive`.
        """
        self.logging_interval = logging_interval
        self.log_on_iteration = log_on_iteration
        if not self.logging_interval and not self.log_on_iteration:
            logger.warning(
                "All logging disabled. Enabling logging on iteration"
            )
            self.log_on_iteration = True
        if self.log_on_iteration:
            if self.logging_interval is None:
                self.logging_interval = self.nlive
            self._last_log = 0
        else:
            self._last_log = time.time()

    @abstractmethod
    def log_state(self):
        raise NotImplementedError()

    def periodically_log_state(self):
        """Log the state of the sampler.

        Calls :code:`log_state` if the elapsed interval in time (or iterations)
        is more than the specified interval.
        """
        if self.log_on_iteration:
            if not (self.iteration - self._last_log) >= self.logging_interval:
                return
            else:
                self._last_log = self.iteration
        else:
            now = time.time()
            if not (now - self._last_log) >= self.logging_interval:
                return
            else:
                self._last_log = now
        self.log_state()

    def initialise_history(self) -> None:
        """Initialise the dictionary to store history"""
        if self.history is None:
            logger.debug("Initialising history dictionary")
            self.history = dict(
                likelihood_evaluations=[],
                sampling_time=[],
                checkpoint_iterations=[],
            )
        else:
            logger.debug("History dictionary already initialised")

    def update_history(self) -> None:
        """Update the history dictionary"""
        self.history["likelihood_evaluations"].append(
            self.total_likelihood_evaluations
        )
        self.history["sampling_time"].append(
            self.current_sampling_time.total_seconds()
        )

    def checkpoint(
        self,
        periodic: bool = False,
        force: bool = False,
        save_existing: bool = True,
    ):
        """Checkpoint the classes internal state.

        Parameters
        ----------
        periodic : bool
            Indicates if the checkpoint is regular periodic checkpointing
            or forced by a signal. If forced by a signal, it will show up on
            the state plot.
        force : bool
            Force the sampler to checkpoint.
        save_existing : bool
            If True, the previous checkpoint will be copied to `.old`.
            If False, the new checkpoint will override the previous file.
        """
        now = datetime.datetime.now()
        if not periodic:
            if self.history:
                self.history["checkpoint_iterations"] += [self.iteration]
            else:
                logger.warning(
                    "Could not log checkpoint iteration in the history"
                )
        elif force:
            pass
        else:
            if self.checkpoint_on_iteration:
                if (
                    self.iteration - self._last_checkpoint
                ) >= self.checkpoint_interval:
                    self._last_checkpoint = self.iteration
                else:
                    return
            else:
                if (
                    now - self._last_checkpoint
                ).total_seconds() >= self.checkpoint_interval:
                    self._last_checkpoint = now
                else:
                    return
        self.sampling_time += now - self.sampling_start_time
        logger.info("Checkpointing nested sampling")
        if self.checkpoint_callback:
            self.checkpoint_callback(self)
        else:
            safe_file_dump(
                self, self.resume_file, pickle, save_existing=save_existing
            )
        self.sampling_start_time = datetime.datetime.now()

    @classmethod
    def resume_from_pickled_sampler(
        cls,
        sampler: Any,
        model: Model,
        output: Optional[str] = None,
        checkpoint_callback: Optional[Callable] = None,
    ):
        """Resume from pickle data.

        Parameters
        ----------
        data : Any
            Pickle data
        model : :obj:`nessai.model.Model`
            User-defined model
        output : Optional[str]
            New output directory. If not specified, the output directory will
            be the same as the previous run.
        checkpoint_callback : Optional[Callable]
            Checkpoint callback function. If not specified, the default method
            will be used.

        Returns
        -------
        Instance of BaseNestedSampler
        """
        logger.info(f"Resuming instance of {cls.__name__}")
        if model.rng is None:
            model.set_rng(sampler.rng)
        model.likelihood_evaluations += (
            sampler._previous_likelihood_evaluations
        )
        model.likelihood_evaluation_time += datetime.timedelta(
            seconds=sampler._previous_likelihood_evaluation_time
        )
        if output is not None and output != sampler.output:
            logger.info(
                f"Overwriting output from {sampler.output} to {output}"
            )
            os.makedirs(output, exist_ok=True)
            sampler.update_output(output)
        sampler.model = model
        sampler.resumed = True
        sampler.checkpoint_callback = checkpoint_callback
        return sampler

    @classmethod
    def resume(
        cls,
        filename: str,
        model: Model,
        output: Optional[str] = None,
        **kwargs,
    ):
        """Resumes the interrupted state from a checkpoint pickle file.

        Parameters
        ----------
        filename : str
            Pickle file to resume from
        model : :obj:`nessai.model.Model`
            User-defined model
        output : Optional[str]
            New output directory. If not specified, the output directory will
            be the same as the previous run.

        Returns
        -------
        obj
            Instance of BaseNestedSampler
        """
        logger.info(f"Resuming {cls.__name__} from {filename}")
        with open(filename, "rb") as f:
            sampler = pickle.load(f)
        return cls.resume_from_pickled_sampler(
            sampler, model, output=output, **kwargs
        )

    @abstractmethod
    def nested_sampling_loop(self):
        raise NotImplementedError()

    def close_pool(self, code=None):
        """Close the multiprocessing pool."""
        self.model.close_pool(code=code)

    def get_result_dictionary(self):
        """Return a dictionary that contains results.

        Only includes version, seed and sampling time. Child classes should
        call this method and add to the dictionary.
        """
        d = dict()
        d["version"] = version
        d["glasflow_version"] = glasflow_version
        d["seed"] = self.seed
        d["sampling_time"] = self.sampling_time.total_seconds()
        d["total_likelihood_evaluations"] = self.model.likelihood_evaluations
        d["likelihood_evaluation_time"] = (
            self.likelihood_evaluation_time.total_seconds()
        )
        d["history"] = self.history
        if hasattr(self.model, "truth"):
            d["truth"] = self.model.truth
        return d

    def __getstate__(self):
        d = self.__dict__
        exclude = {"model", "proposal", "checkpoint_callback"}
        state = {k: d[k] for k in d.keys() - exclude}
        state["_previous_likelihood_evaluations"] = d[
            "model"
        ].likelihood_evaluations
        state["_previous_likelihood_evaluation_time"] = d[
            "model"
        ].likelihood_evaluation_time.total_seconds()
        return state



"""General configuration for the test suite"""

import logging
import multiprocessing
import sys
import time

import numpy as np
import pytest
import torch
from scipy.stats import norm

from nessai.livepoint import (
    add_extra_parameters_to_live_points,
    reset_extra_live_points_parameters,
)
from nessai.model import Model
from nessai.utils.testing import IntegrationTestModel

torch.manual_seed(170817)

_requires_dependency_cache = dict()


@pytest.fixture()
def rng():
    return np.random.default_rng(170817)


@pytest.fixture()
def model(rng):
    class TestModel(Model):
        def __init__(self):
            self.bounds = {"x": [-5, 5], "y": [-5, 5]}
            self.names = ["x", "y"]

        def log_prior(self, x):
            log_p = np.log(self.in_bounds(x), dtype="float")
            for n in self.names:
                log_p -= self.bounds[n][1] - self.bounds[n][0]
            return log_p

        def log_likelihood(self, x):
            log_l = 0
            for pn in self.names:
                log_l += norm.logpdf(x[pn])
            return log_l

        def to_unit_hypercube(self, x):
            x_out = x.copy()
            for n in self.names:
                x_out[n] = (x[n] - self.bounds[n][0]) / np.ptp(self.bounds[n])
            return x_out

        def from_unit_hypercube(self, x):
            x_out = x.copy()
            for n in self.names:
                x_out[n] = np.ptp(self.bounds[n]) * x[n] + self.bounds[n][0]
            return x_out

    m = TestModel()
    m.set_rng(rng)
    return m


@pytest.fixture(scope="function")
def integration_model(rng):
    model = IntegrationTestModel()
    return model


@pytest.fixture()
def flow_config():
    return dict(
        n_blocks=2,
        n_neurons=2,
        n_layers=1,
        batch_norm_between_layers=False,
    )


@pytest.fixture()
def training_config():
    return dict(
        max_epochs=5,
        device_tag="cpu",
    )


@pytest.fixture()
def wait():
    """Sleep for 0.01s for timing tests.

    Time resolution on Windows is less than Linux and macOS
    """

    def func(*args, **kwargs):
        time.sleep(0.01)

    return func


@pytest.fixture(params=["fork", "forkserver", "spawn"])
def mp_context(request):
    """Multiprocessing context to test"""
    if sys.platform == "win32" and request.param != "spawn":
        pytest.skip("Windows only supports the 'spawn' start method")
    return multiprocessing.get_context(request.param)


@pytest.fixture()
def ins_parameters():
    """Add (and remove) the standard INS parameters for the tests."""
    # Before every test
    add_extra_parameters_to_live_points(["logQ", "logW", "logU"])
    yield
    reset_extra_live_points_parameters()


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers", "cuda: mark test to indicate it requires CUDA"
    )
    config.addinivalue_line(
        "markers",
        "requires(package): mark test to only run if the package can be "
        "imported",
    )
    config.addinivalue_line(
        "markers",
        "skip_on_windows: mark test to indicated it should be skipped on "
        "Windows",
    )
    config.addinivalue_line(
        "markers",
        "bilby_compatibility: mark test as a bilby compatibility test, these "
        "tests will be skipped unless the command line option is specified.",
    )
    config.addinivalue_line(
        "markers",
        "reset_logger: reset the logger after the test",
    )


def pytest_addoption(parser):
    parser.addoption(
        "--bilby-compatibility",
        action="store_true",
        default=False,
        help="Run bilby compatibility tests",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--bilby-compatibility"):
        return
    skip_bilby_compat = pytest.mark.skip(
        reason="Need --bilby-compatibility to run"
    )
    for item in items:
        if "bilby_compatibility" in item.keywords:
            item.add_marker(skip_bilby_compat)


def pytest_runtest_setup(item):
    for mark in item.iter_markers(name="cuda"):
        if not torch.cuda.is_available():
            pytest.skip("Test requires CUDA")
    for mark in item.iter_markers(name="requires"):
        name = mark.args[0]
        if name in _requires_dependency_cache:
            skip_it = _requires_dependency_cache[name]
        else:
            try:
                __import__(name)
                skip_it = False
            except ImportError:
                skip_it = True

            _requires_dependency_cache[name] = skip_it

        reason = "Missing dependency: {}".format(name)
        if skip_it:
            pytest.skip(reason)
    for mark in item.iter_markers(name="skip_on_windows"):
        if sys.platform == "win32":
            pytest.skip("Test does not run on Windows")


def pytest_runtest_teardown(item):
    if "reset_logger" in item.keywords:
        logger = logging.getLogger("nessai")
        logger.setLevel(logging.NOTSET)
        logger.propagate = True
        logger.disabled = False
        logger.filters.clear()
        handlers = logger.handlers.copy()
        for handler in handlers:
            # Copied from `logging.shutdown`.
            try:
                handler.acquire()
                handler.flush()
                handler.close()
            except (OSError, ValueError):
                pass
            finally:
                handler.release()
            logger.removeHandler(handler)
        logger.addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
"""
Functions and classes for using nessai for gravitational-wave inference.
"""

import warnings

from .proposal import GWFlowProposal

__all__ = ["GWFlowProposal"]

warnings.warn(
    (
        "The `nessai.gw` module will be deprecated in the next release in "
        "favour of the nessai-gw package. This packages provides the same "
        "functionality as`nessai.gw` via the plugin interface."
        "For more details, see: https://github.com/mj-will/nessai-gw"
    ),
    FutureWarning,
)



from unittest.mock import patch

import pytest

from nessai.proposal.flowproposal.flowproposal import FlowProposal


@pytest.mark.parametrize(
    "expansion_fraction, fuzz", [(None, 2.0), (0.5, 1.5**0.5)]
)
def test_set_rescaling(proposal, expansion_fraction, fuzz):
    proposal.rescaled_dims = 2
    proposal.expansion_fraction = expansion_fraction
    proposal.fuzz = 2.0
    with patch(
        "nessai.proposal.flowproposal.flowproposal.BaseFlowProposal.set_rescaling"  # noqa: E501
    ) as mock_parent:
        FlowProposal.set_rescaling(proposal)
    mock_parent.assert_called_once()
    proposal.configure_constant_volume.assert_called_once()
    assert proposal.fuzz == fuzz



from inspect import signature

import pytest

from nessai.reparameterisations import (
    default_reparameterisations,
)


@pytest.fixture(params=default_reparameterisations.values())
def reparameterisation(request):
    """Fixture to test all reparameterisations."""
    return request.param


def test_reparams_rng(reparameterisation):
    """Test that all reparameterisations can be initialised with an rng"""
    sig = signature(reparameterisation.class_fn)
    assert "rng" in sig.parameters



import numpy as np


def next_pow_two(n):
    """Returns the next power of two greater than or equal to `n`

    Copied from emcee.
    """
    i = 1
    while i < n:
        i = i << 1
    return i


def function_1d(x):
    """Estimate the normalized autocorrelation function of a 1-D series

    Copied from emcee
    """
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")
    n = next_pow_two(len(x))

    # Compute the FFT and then (from that) the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
    if acf[0] == 0.0:
        acf[:] = np.nan_to_num(np.inf)
    else:
        acf /= acf[0]
    return acf


def auto_window(taus, c):
    m = np.arange(len(taus)) < c * taus
    if np.any(m):
        return np.argmin(m)
    return len(taus) - 1


def integrated_time(x, c=5):
    """Estimate the integrated autocorrelation time of a time series.

    Based on the implementation from emcee.
    """
    n_t, n_w, n_d = x.shape
    tau_est = np.empty(n_d)
    windows = np.empty(n_d, dtype=int)

    # Loop over parameters
    for d in range(n_d):
        f = np.zeros(n_t)
        for k in range(n_w):
            f += function_1d(x[:, k, d])
        f /= n_w
        taus = 2.0 * np.cumsum(f) - 1.0
        windows[d] = auto_window(taus, c)
        tau_est[d] = taus[windows[d]]

    return tau_est



import logging
import math
from functools import lru_cache
from typing import Optional, Tuple

import numpy as np

logger = logging.getLogger(__name__)


class Step:
    def __init__(
        self,
        dims: int,
        ensemble: Optional[np.ndarray] = None,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        self.dims = dims
        self.update_ensemble(ensemble)
        self.rng = rng or np.random.default_rng()

    def step(self, z):
        raise NotImplementedError

    def update_ensemble(self, ensemble):
        self.ensemble = ensemble

    def update_stats(self, n_accept, n_reject):
        self.n_accept = n_accept
        self.n_reject = n_reject

    def __call__(self, *args, **kwargs):
        return self.step(*args, **kwargs)


class GaussianStep(Step):
    def __init__(
        self,
        dims: int,
        ensemble: Optional[np.ndarray] = None,
        rng: Optional[np.random.Generator] = None,
        scale: Optional[float] = None,
        update_scale: bool = True,
        target_acceptance: float = 0.5,
    ) -> None:
        super().__init__(dims=dims, ensemble=ensemble, rng=rng)
        if scale is None:
            self.scale = 2 / self.dims**0.5
        else:
            self.scale = scale
        self.update_scale = update_scale
        self.target_acceptance = target_acceptance

    def step(self, z):
        return self.rng.normal(z, scale=self.scale, size=z.shape), np.zeros(
            z.shape[0]
        )

    def update_stats(self, n_accept, n_reject):
        super().update_stats(n_accept, n_reject)
        if self.update_scale:
            # This follows the same maths as dynesty
            acceptance = self.n_accept / (self.n_accept + self.n_reject)
            logger.debug(f"Current acceptance: {acceptance}")
            self.scale *= math.exp(
                (acceptance - self.target_acceptance)
                / self.dims
                / self.target_acceptance
            )
            logger.debug(f"New scale: {self.scale}")


@lru_cache(maxsize=1)
def _get_nondiagonal_pairs(n: int) -> np.ndarray:
    """Get the indices of a square matrix with size n, excluding the diagonal.

    This is direct copy the same function from emcee.
    """
    rows, cols = np.tril_indices(n, -1)  # -1 to exclude diagonal

    # Combine rows-cols and cols-rows pairs
    pairs = np.column_stack(
        [np.concatenate([rows, cols]), np.concatenate([cols, rows])]
    )
    return pairs


class DifferentialEvolutionStep(Step):
    """Differential evolution step for MCMC proposal.

    Based on the differential evolution proposal from CPnest and the dynesty-
    specific implementation in bilby.

    Parameters
    ----------
    dims : int
        The number of dimensions in the parameter space.
    ensemble : np.ndarray, optional
        The ensemble of live points. Default is None.
    mix_fraction : float, optional
        The fraction of the ensemble to mix. Default is 0.5.
    sigma : float, optional
        The standard deviation of the error term added to the steps.
        Default is 1e-5.
    """

    def __init__(
        self,
        dims: int,
        ensemble: Optional[np.ndarray] = None,
        rng: Optional[np.random.Generator] = None,
        mix_fraction: float = 0.5,
        sigma: float = 1e-5,
    ) -> None:
        self.pairs = None
        self.n_pairs = None

        super().__init__(dims=dims, ensemble=ensemble, rng=rng)

        self.g0 = 2.38 / np.sqrt(2 * self.dims)
        self.mix_fraction = mix_fraction
        self.sigma = sigma

    def update_ensemble(self, ensemble):
        super().update_ensemble(ensemble)
        if self.ensemble is not None:
            self.pairs = _get_nondiagonal_pairs(self.ensemble.shape[0])
            self.n_pairs = self.pairs.shape[0]

    def step(self, z):
        nc = z.shape[0]
        indices = self.rng.choice(self.n_pairs, size=nc, replace=True)
        diffs = np.diff(self.ensemble[self.pairs[indices]], axis=1).squeeze(
            axis=1
        )
        mix = self.rng.random(nc) < self.mix_fraction
        scale = np.ones((nc, 1))
        scale[mix, :] = self.g0
        error = self.sigma * self.rng.standard_normal((nc, self.dims))
        z_new = z + scale * diffs + error
        return z_new, np.zeros(nc)


class StretchStep(Step):
    """Stretch step for MCMC proposal.

    Based on the stretch move proposal from Goodman & Weare (2010), as
    implemented in :code:`cpnest` and :code:`LALInference`.

    Parameters
    ----------
    dims : int
        The number of dimensions in the parameter space.
    scale : float, optional
        The scale factor for the stretch move. Default is 2.0.
    ensemble : np.ndarray, optional
        The ensemble of live points. Default is None.
    """

    def __init__(
        self,
        dims: int,
        ensemble: np.ndarray = None,
        rng: Optional[np.random.Generator] = None,
        scale: float = 2.0,
    ):
        super().__init__(dims=dims, ensemble=ensemble, rng=rng)
        self.scale = scale

    def step(self, z: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        nc = z.shape[0]
        indices = self.rng.choice(len(self.ensemble), size=nc, replace=True)
        a = self.ensemble[indices]
        x = self.rng.uniform(-1, 1, size=nc) * np.log(self.scale)
        scale = np.exp(x)
        z_new = a + (z - a) * scale[:, None]
        return z_new, x * self.dims


KNOWN_STEPS = dict(
    gaussian=GaussianStep,
    diff=DifferentialEvolutionStep,
    stretch=StretchStep,
)



# -*- coding: utf-8 -*-
"""
Functions and objects related to reparametersiations for use in the
``reparameterisations`` dictionary.

See the documentation for an in-depth description of how to use these
functions and classes.
"""

import logging

from .angle import Angle, AnglePair, ToCartesian
from .base import Reparameterisation
from .combined import CombinedReparameterisation
from .discrete import Dequantise
from .null import NullReparameterisation
from .rescale import Rescale, RescaleToBounds, ScaleAndShift
from .utils import (
    KnownReparameterisation,
    ReparameterisationDict,
    ReparameterisationError,
    get_reparameterisation,
)

logger = logging.getLogger(__name__)


default_reparameterisations = ReparameterisationDict()
default_reparameterisations.add_reparameterisation("default", RescaleToBounds)
default_reparameterisations.add_reparameterisation(
    "rescaletobounds", RescaleToBounds
)
default_reparameterisations.add_reparameterisation(
    "rescale-to-bounds", RescaleToBounds
)

default_reparameterisations.add_reparameterisation(
    "offset",
    RescaleToBounds,
    {"offset": True},
)
default_reparameterisations.add_reparameterisation(
    "inversion",
    RescaleToBounds,
    {
        "detect_edges": True,
        "boundary_inversion": True,
        "inversion_type": "split",
    },
)
default_reparameterisations.add_reparameterisation(
    "inversion-duplicate",
    RescaleToBounds,
    {
        "detect_edges": True,
        "boundary_inversion": True,
        "inversion_type": "duplicate",
    },
)
default_reparameterisations.add_reparameterisation(
    "logit",
    RescaleToBounds,
    {
        "rescale_bounds": [0.0, 1.0],
        "update_bounds": False,
        "post_rescaling": "logit",
    },
)
default_reparameterisations.add_reparameterisation(
    "log-rescale",
    RescaleToBounds,
    {
        "rescale_bounds": [0.0, 1.0],
        "update_bounds": False,
        "post_rescaling": "log",
    },
)
default_reparameterisations.add_reparameterisation("scale", Rescale)
default_reparameterisations.add_reparameterisation(
    "scaleandshift", ScaleAndShift
)
default_reparameterisations.add_reparameterisation("rescale", Rescale)
default_reparameterisations.add_reparameterisation(
    "zscore",
    ScaleAndShift,
    {"estimate_scale": True, "estimate_shift": True},
)
default_reparameterisations.add_reparameterisation(
    "standardize",
    ScaleAndShift,
    {"estimate_scale": True, "estimate_shift": True},
)
default_reparameterisations.add_reparameterisation(
    "z-score",
    ScaleAndShift,
    {"estimate_scale": True, "estimate_shift": True},
)
default_reparameterisations.add_reparameterisation(
    "zscore-gaussian-cdf",
    ScaleAndShift,
    {
        "estimate_scale": True,
        "estimate_shift": True,
        "post_rescaling": "gaussian_cdf",
    },
)
default_reparameterisations.add_reparameterisation(
    "z-score-gaussian-cdf",
    ScaleAndShift,
    {
        "estimate_scale": True,
        "estimate_shift": True,
        "post_rescaling": "gaussian_cdf",
    },
)
default_reparameterisations.add_reparameterisation(
    "z-score-logit",
    ScaleAndShift,
    {
        "estimate_scale": True,
        "estimate_shift": True,
        "pre_rescaling": "logit",
    },
)
default_reparameterisations.add_reparameterisation(
    "zscore-logit",
    ScaleAndShift,
    {
        "estimate_scale": True,
        "estimate_shift": True,
        "pre_rescaling": "logit",
    },
)
default_reparameterisations.add_reparameterisation(
    "z-score-inv-gaussian-cdf",
    ScaleAndShift,
    {
        "estimate_scale": True,
        "estimate_shift": True,
        "pre_rescaling": "inv_gaussian_cdf",
    },
)
default_reparameterisations.add_reparameterisation(
    "zscore-inv-gaussian-cdf",
    ScaleAndShift,
    {
        "estimate_scale": True,
        "estimate_shift": True,
        "pre_rescaling": "inv_gaussian_cdf",
    },
)
default_reparameterisations.add_reparameterisation(
    "log-z-score",
    ScaleAndShift,
    {"estimate_scale": True, "estimate_shift": True, "pre_rescaling": "log"},
)
default_reparameterisations.add_reparameterisation(
    "log-standardise",
    ScaleAndShift,
    {"estimate_scale": True, "estimate_shift": True, "pre_rescaling": "log"},
)
default_reparameterisations.add_reparameterisation("angle", Angle, {})
default_reparameterisations.add_reparameterisation(
    "angle-pi", Angle, {"scale": 2.0, "prior": "uniform"}
)
default_reparameterisations.add_reparameterisation(
    "angle-2pi", Angle, {"scale": 1.0, "prior": "uniform"}
)
default_reparameterisations.add_reparameterisation(
    "angle-sine", RescaleToBounds
)
default_reparameterisations.add_reparameterisation(
    "angle-cosine", RescaleToBounds
)
default_reparameterisations.add_reparameterisation("angle-pair", AnglePair)
default_reparameterisations.add_reparameterisation(
    "periodic", Angle, {"scale": None}
)
default_reparameterisations.add_reparameterisation("to-cartesian", ToCartesian)
default_reparameterisations.add_reparameterisation("dequantise", Dequantise)
default_reparameterisations.add_reparameterisation(
    "dequantise-logit",
    Dequantise,
    {
        "rescale_bounds": [0.0, 1.0],
        "update_bounds": False,
        "post_rescaling": "logit",
    },
)
default_reparameterisations.add_reparameterisation(
    "none", NullReparameterisation
)
default_reparameterisations.add_reparameterisation(
    "null", NullReparameterisation
)
default_reparameterisations.add_reparameterisation(
    None, NullReparameterisation
)

default_reparameterisations.add_external_reparameterisations(
    "nessai.reparameterisations"
)


__all__ = [
    "Angle",
    "AnglePair",
    "CombinedReparameterisation",
    "Dequantise",
    "KnownReparameterisation",
    "NullReparameterisation",
    "Reparameterisation",
    "ReparameterisationError",
    "Rescale",
    "RescaleToBounds",
    "ToCartesian",
    "get_reparameterisation",
]



# -*- coding: utf-8 -*-
"""Test general configuration functions"""

from unittest.mock import patch

import pytest

from nessai import utils
from nessai.proposal import FlowProposal


def test_config_drawsize_none(proposal):
    """Test the popluation configuration with no drawsize given"""
    proposal.poolsize = 2000
    FlowProposal.configure_population(proposal, None, 1.0, 0.0, "gaussian")
    assert proposal.drawsize == 2000


def test_configure_latent_temperature(proposal):
    """Test the configuration of the latent temperature"""
    FlowProposal.configure_population(
        proposal, 1000, 1.0, 0.0, "gaussian", 0.9
    )
    assert proposal.latent_temperature == 0.9


def test_configure_latent_temperature_invalid(proposal):
    """Test the configuration of the latent temperature"""
    with pytest.raises(
        ValueError,
        match="Latent temperature can only be used with a Gaussian latent",
    ):
        FlowProposal.configure_population(
            proposal, 1000, 1.0, 0.0, "truncated_gaussian", 0.9
        )


@pytest.mark.parametrize("fixed_radius", [False, 5.0, 1])
def test_config_fixed_radius(proposal, fixed_radius):
    """Test the configuration for a fixed radius"""
    FlowProposal.configure_fixed_radius(proposal, fixed_radius)
    assert proposal.fixed_radius == fixed_radius


def test_config_fixed_radius_not_float(proposal):
    """
    Test the fixed radius is disabled when the radius cannot be converted to
    a float.
    """
    FlowProposal.configure_fixed_radius(proposal, "four")
    assert proposal.fixed_radius is False


def test_min_radius_no_max(proposal):
    """Test configuration of min radius and no max radius"""
    FlowProposal.configure_min_max_radius(proposal, 5.0, False)
    assert proposal.min_radius == 5.0
    assert proposal.max_radius is False


def test_min_max_radius(proposal):
    """Test configuration of min radius and no max radius"""
    FlowProposal.configure_min_max_radius(proposal, 5, 10)
    assert proposal.min_radius == 5.0
    assert proposal.max_radius == 10.0


@pytest.mark.parametrize("rmin, rmax", [(None, 1.0), (1.0, "2")])
def test_min_max_radius_invalid_input(proposal, rmin, rmax):
    """Test configuration of min radius and no max radius"""
    with pytest.raises(RuntimeError):
        FlowProposal.configure_min_max_radius(proposal, rmin, rmax)


@pytest.mark.parametrize(
    "latent_prior, prior_func",
    [
        ("gaussian", "draw_gaussian"),
        ("truncated_gaussian", "draw_truncated_gaussian"),
        ("uniform", "draw_uniform"),
        ("uniform_nsphere", "draw_nsphere"),
        ("uniform_nball", "draw_nsphere"),
        ("flow", None),
    ],
)
def test_configure_latent_prior(proposal, latent_prior, prior_func):
    """Test to make sure the correct latent priors are used."""
    proposal.latent_prior = latent_prior
    proposal.flow_config = {"model_config": {}}
    FlowProposal.configure_latent_prior(proposal)
    if prior_func:
        assert proposal._draw_latent_prior == getattr(utils, prior_func)
    else:
        assert proposal._draw_latent_prior is None


def test_configure_latent_prior_unknown(proposal):
    """Make sure unknown latent priors raise an error"""
    proposal.latent_prior = "truncated"
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_latent_prior(proposal)
    assert "Unknown latent prior: truncated, " in str(excinfo.value)


@pytest.mark.parametrize(
    "latent_prior",
    ["truncated_gaussian", "uniform_nball", "uniform_nsphere"],
)
def test_configure_constant_volume(proposal, latent_prior):
    """Test configuration for constant volume mode."""
    proposal.constant_volume_mode = True
    proposal.volume_fraction = 0.95
    proposal.rescaled_dims = 5
    proposal.latent_prior = latent_prior
    proposal.max_radius = 3.0
    proposal.min_radius = 5.0
    proposal.fuzz = 1.5
    with patch(
        "nessai.proposal.flowproposal.flowproposal.compute_radius",
        return_value=4.0,
    ) as mock:
        FlowProposal.configure_constant_volume(proposal)
    mock.assert_called_once_with(5, 0.95)
    assert proposal.fixed_radius == 4.0
    assert proposal.min_radius is False
    assert proposal.max_radius is False
    assert proposal.fuzz == 1.0


def test_configure_constant_volume_disabled(proposal):
    """Assert nothing happens if constant_volume is False"""
    proposal.constant_volume_mode = False
    with patch(
        "nessai.proposal.flowproposal.flowproposal.compute_radius"
    ) as mock:
        FlowProposal.configure_constant_volume(proposal)
    mock.assert_not_called()


def test_constant_volume_invalid_latent_prior(proposal):
    """Assert an error is raised if the latent prior is not a truncated \
        Gaussian
    """
    err = "Constant volume mode is not supported for latent_prior=gaussian"
    proposal.constant_volume_mode = True
    proposal.latent_prior = "gaussian"
    with pytest.raises(RuntimeError, match=err):
        FlowProposal.configure_constant_volume(proposal)



import sys

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


def get_entry_points(group: str) -> dict:
    """Return a dictionary of entry points for a given group

    Parameters
    ----------
    group: str
        Entry points you wish to query

    Returns
    -------
    dict
        A dictionary containing the entry points.
    """
    return {custom.name: custom for custom in entry_points(group=group)}



"""Proposal class that uses normalising flows and rejection sampling.

This is the default proposal used by nessai.
"""

import datetime
import logging
from functools import partial

import numpy as np
from scipy.special import logsumexp

from ... import config
from ...livepoint import (
    empty_structured_array,
    numpy_array_to_live_points,
)
from ...utils import (
    compute_radius,
    get_uniform_distribution,
)
from ...utils.sampling import NDimensionalTruncatedGaussian
from ...utils.structures import get_subset_arrays
from .base import BaseFlowProposal

logger = logging.getLogger(__name__)


class FlowProposal(BaseFlowProposal):
    """
    Object that handles training and proposal points

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User defined model.
    latent_prior : {'truncated_gaussian', 'gaussian', 'uniform_nsphere', \
            'gaussian'}, optional
        Prior distribution in the latent space. Defaults to
        'truncated_gaussian'.
    poolsize : int, optional
        Size of the proposal pool. Defaults to 10000.
    drawsize : int, optional
        Number of points to simultaneously draw when populating the proposal
        Defaults to 10000
    min_radius : float, optional
        Minimum radius used for population. If not specified not minimum is
        used.
    max_radius : float, optional
        If a float then this value is used as an upper limit for the
        computed radius when populating the proposal. If unspecified no
        upper limit is used.
    fixed_radius : float, optional
        If specified and the chosen latent distribution is compatible, this
        radius will be used to draw new samples instead of the value computed
        with the flow.
    constant_volume_mode : bool
        If True, then a constant volume is used for the latent contour used to
        draw new samples. The exact volume can be set using `volume_fraction`
    volume_fraction : float
        Fraction of the total probability to contain with the latent contour
        when using a constant volume.
    compute_radius_with_all : bool, optional
        If True all the radius of the latent contour is computed using the
        maximum radius of all the samples used to train the flow.
    fuzz : float, optional
        Fuzz-factor applied to the radius. If unspecified no fuzz-factor is
        applied.
    expansion_fraction : float, optional
        Similar to ``fuzz`` but instead a scaling factor applied to the radius
        this specifies a rescaling for volume of the n-ball used to draw
        samples. This is translated to a value for ``fuzz``.
    truncate_log_q : bool, optional
        Truncate proposals using minimum log-probability of the training data.
    enforce_likelihood_threshold : bool
        If True, enforce the likelihood threshold when performing rejection
        sampling. If false, the likelihood is not checked when populating the
        proposal.
    """

    def __init__(
        self,
        model,
        poolsize=None,
        latent_prior="truncated_gaussian",
        latent_temperature=None,
        constant_volume_mode=True,
        volume_fraction=0.95,
        fuzz=1.0,
        fixed_radius=False,
        drawsize=None,
        truncate_log_q=False,
        expansion_fraction=4.0,
        min_radius=False,
        max_radius=50.0,
        compute_radius_with_all=False,
        enforce_likelihood_threshold=False,
        **kwargs,
    ):
        super().__init__(
            model,
            poolsize=poolsize,
            **kwargs,
        )
        logger.debug("Initialising FlowProposal")

        self._draw_func = None
        self._populate_dist = None

        self.configure_population(
            drawsize,
            fuzz,
            expansion_fraction,
            latent_prior,
            latent_temperature,
        )

        self.truncate_log_q = truncate_log_q
        self.constant_volume_mode = constant_volume_mode
        self.volume_fraction = volume_fraction
        self.enforce_likelihood_threshold = enforce_likelihood_threshold

        self.compute_radius_with_all = compute_radius_with_all
        self.configure_fixed_radius(fixed_radius)
        self.configure_min_max_radius(min_radius, max_radius)

        self.configure_latent_prior()
        self.alt_dist = None

    def configure_population(
        self,
        drawsize,
        fuzz,
        expansion_fraction,
        latent_prior,
        latent_temperature=None,
    ):
        """
        Configure settings related to population
        """
        if drawsize is None:
            drawsize = self.poolsize

        self.drawsize = drawsize
        self.fuzz = fuzz
        self.expansion_fraction = expansion_fraction
        self.latent_prior = latent_prior
        if latent_temperature is not None:
            if latent_prior != "gaussian":
                raise ValueError(
                    "Latent temperature can only be used with a Gaussian latent prior"
                )
            else:
                logger.warning("`latent_temperature` is experimental!")
        self.latent_temperature = latent_temperature

    def configure_latent_prior(self):
        """Configure the latent prior"""
        if self.latent_prior == "truncated_gaussian":
            from ...utils import draw_truncated_gaussian

            self._draw_latent_prior = draw_truncated_gaussian

        elif self.latent_prior == "gaussian":
            logger.warning("Using a gaussian latent prior WITHOUT truncation")
            from ...utils import draw_gaussian

            self._draw_latent_prior = draw_gaussian
        elif self.latent_prior == "uniform":
            from ...utils import draw_uniform

            self._draw_latent_prior = draw_uniform
        elif self.latent_prior in ["uniform_nsphere", "uniform_nball"]:
            from ...utils import draw_nsphere

            self._draw_latent_prior = draw_nsphere
        elif self.latent_prior == "flow":
            self._draw_latent_prior = None
        else:
            raise RuntimeError(
                f"Unknown latent prior: {self.latent_prior}, choose from: "
                "truncated_gaussian (default), gaussian, "
                "uniform, uniform_nsphere"
            )

    def configure_fixed_radius(self, fixed_radius):
        """Configure the fixed radius"""
        if fixed_radius:
            try:
                self.fixed_radius = float(fixed_radius)
            except ValueError:
                logger.error(
                    "Fixed radius enabled but could not be converted to a "
                    "float. Setting fixed_radius=False"
                )
                self.fixed_radius = False
        else:
            self.fixed_radius = False

    def configure_min_max_radius(self, min_radius, max_radius):
        """
        Configure the minimum and maximum radius
        """
        if isinstance(min_radius, (int, float)):
            self.min_radius = float(min_radius)
        else:
            raise RuntimeError("Min radius must be an int or float")

        if max_radius:
            if isinstance(max_radius, (int, float)):
                self.max_radius = float(max_radius)
            else:
                raise RuntimeError("Max radius must be an int or float")
        else:
            logger.warning(
                "Running without a maximum radius! The proposal "
                "process may get stuck if very large radii are "
                "returned by the worst point."
            )
            self.max_radius = False

    def configure_constant_volume(self):
        """Configure using constant volume latent contour."""
        if self.constant_volume_mode:
            logger.debug("Configuring constant volume latent contour")
            if self.latent_prior == "truncated_gaussian":
                pass
            elif self.latent_prior in ["uniform_nball", "uniform_nsphere"]:
                logger.warning(
                    "Constant volume mode with latent_prior="
                    f"{self.latent_prior} is experimental!"
                )
            else:
                raise RuntimeError(
                    "Constant volume mode is not supported for latent_prior="
                    f"{self.latent_prior}"
                )
            self.fixed_radius = compute_radius(
                self.rescaled_dims, self.volume_fraction
            )
            self.fuzz = 1.0
            if self.max_radius < self.fixed_radius:
                logger.warning(
                    "Max radius is less than the radius need to use a "
                    "constant volume latent contour. Max radius will be "
                    "disabled."
                )
                self.max_radius = False
            if self.min_radius > self.fixed_radius:
                logger.warning(
                    "Min radius is greater than the radius need to use a "
                    "constant volume latent contour. Min radius will be "
                    "disabled."
                )
                self.min_radius = False
        else:
            logger.debug(
                "Nothing to configure for constant volume latent contour."
            )

    def set_rescaling(self):
        """Set the rescaling functions.

        Also configures constant volume mode since this must be done AFTER
        the reparameterisations are configured.
        """
        super().set_rescaling()
        if self.expansion_fraction and self.expansion_fraction is not None:
            logger.info("Overwriting fuzz factor with expansion fraction")
            self.fuzz = (1 + self.expansion_fraction) ** (
                1 / self.rescaled_dims
            )
            logger.info(f"New fuzz factor: {self.fuzz}")
        self.configure_constant_volume()

    def radius(self, z, *arrays):
        """
        Calculate the radius of a latent point or set of latent points.
        If multiple points are parsed the maximum radius is returned.

        Parameters
        ----------
        z : :obj:`np.ndarray`
            Array of points in the latent space
        *arrays :
            Additional arrays to return the corresponding value

        Returns
        -------
        tuple of arrays
            Tuple of array with the maximum radius and corresponding values
            from any additional arrays that were passed.
        """
        r = np.sqrt(np.sum(z**2.0, axis=-1))
        i = np.nanargmax(r)
        if arrays:
            return (r[i],) + tuple(a[i] for a in arrays)
        else:
            return r[i]

    def prep_latent_prior(self):
        """Prepare the latent prior."""
        if self.latent_prior == "truncated_gaussian":
            self._populate_dist = NDimensionalTruncatedGaussian(
                self.dims,
                self.r,
                fuzz=self.fuzz,
                rng=self.rng,
            )
            self._draw_func = self._populate_dist.sample
        elif self.latent_prior == "flow":
            self._draw_func = lambda N: self.flow.sample_latent_distribution(N)
        else:
            draw_kwargs = dict(
                dims=self.dims,
                r=self.r,
                fuzz=self.fuzz,
                rng=self.rng,
            )
            if self.latent_temperature is not None:
                draw_kwargs["temperature"] = self.latent_temperature
            assert self.rng is not None
            self._draw_func = partial(
                self._draw_latent_prior,
                **draw_kwargs,
            )

    def draw_latent_prior(self, n):
        """Draw n samples from the latent prior."""
        return self._draw_func(N=n)

    def backward_pass(
        self,
        z,
        rescale=True,
        discard_nans=True,
        return_z=False,
        return_unit_hypercube=False,
    ):
        """
        A backwards pass from the model (latent -> real)

        Parameters
        ----------
        z : array_like
            Structured array of points in the latent space
        rescale : bool, optional (True)
            Apply inverse rescaling function
        discard_nan: bool
            If True, samples with NaNs or Infs in log_q are removed.
        return_z : bool
            If True, return the array of latent samples, this may differ from
            the input since samples can be discarded.

        Returns
        -------
        x : array_like
            Samples in the data space
        log_prob : array_like
            Log probabilities corresponding to each sample (including the
            Jacobian)
        z : array_like
            Samples in the latent space, only returned if :code:`return_z=True`
        """
        # Compute the log probability
        try:
            x, log_prob = self.flow.sample_and_log_prob(
                z=z, alt_dist=self.alt_dist
            )
        except AssertionError as e:
            logger.warning(
                "Assertion error raised when sampling from the flow."
                f"Error: {e}"
            )
            if return_z:
                return np.array([]), np.array([]), np.array([])
            else:
                return np.array([]), np.array([])

        if discard_nans:
            valid = np.isfinite(log_prob)
            x, log_prob, z = get_subset_arrays(valid, x, log_prob, z)
        x = numpy_array_to_live_points(
            x.astype(config.livepoints.default_float_dtype),
            self.prime_parameters,
        )
        # Apply rescaling in rescale=True
        if rescale:
            x, log_J = self.inverse_rescale(
                x, return_unit_hypercube=return_unit_hypercube
            )
            # Include Jacobian for the rescaling
            log_prob -= log_J
            if not return_unit_hypercube:
                x, z, log_prob = self.check_prior_bounds(x, z, log_prob)
        if return_z:
            return x, log_prob, z
        else:
            return x, log_prob

    def populate(
        self,
        worst_point,
        n_samples=10000,
        plot=True,
        r=None,
        max_samples=1_000_000,
    ):
        """
        Populate a pool of latent points given the current worst point.

        Parameters
        ----------
        worst_point : structured_array
            The current worst point used to compute the radius of the contour
            in the latent space.
        n_samples : int, optional (10000)
            The total number of points to populate in the pool
        plot : {True, False, 'all'}
            Enable or disable plots for during training. By default the plots
            are only one-dimensional histograms, `'all'` includes corner
            plots with samples, these are often a few MB in size so
            proceed with caution!
        """
        st = datetime.datetime.now()
        if not self.initialised:
            raise RuntimeError(
                "Proposal has not been initialised. "
                "Try calling `initialise()` first."
            )
        if r is not None:
            logger.debug(f"Using user inputs for radius {r}")
        elif self.fixed_radius:
            r = self.fixed_radius
        else:
            logger.debug(f"Populating with worst point: {worst_point}")
            if self.compute_radius_with_all:
                logger.debug("Using previous live points to compute radius")
                worst_point = self.training_data
            worst_z = self.forward_pass(
                worst_point, rescale=True, compute_radius=True
            )[0]
            r = self.radius(worst_z)
            if self.max_radius and r > self.max_radius:
                r = self.max_radius
            if self.min_radius and r < self.min_radius:
                r = self.min_radius

        if self.truncate_log_q:
            log_q_live_points = self.forward_pass(self.training_data)[1]
            min_log_q = log_q_live_points.min()
            logger.debug(f"Truncating with log_q={min_log_q:.3f}")
        else:
            min_log_q = None

        logger.debug(f"Populating proposal with latent radius: {r:.5}")
        # Radius is not used for the flow or Gaussian latent priors
        if self.latent_prior in ["flow", "gaussian"]:
            self.r = np.nan
        else:
            self.r = r

        self.alt_dist = self.get_alt_distribution()

        if self.indices:
            logger.debug(
                "Existing pool of samples is not empty. "
                "Discarding existing samples."
            )
        self.indices = []

        if self.accumulate_weights:
            samples = empty_structured_array(0, dtype=self.population_dtype)
        else:
            samples = empty_structured_array(
                n_samples, dtype=self.population_dtype
            )

        self.prep_latent_prior()

        log_n = np.log(n_samples)
        log_n_expected = -np.inf
        n_proposed = 0
        log_weights = np.empty(0)
        log_constant = -np.inf
        n_accepted = 0
        accept = None
        log_likelihood_threshold = worst_point["logL"]

        while n_accepted < n_samples:
            z = self.draw_latent_prior(self.drawsize)
            n_proposed += z.shape[0]

            x, log_q = self.backward_pass(
                z,
                rescale=not self.use_x_prime_prior,
                return_unit_hypercube=self.map_to_unit_hypercube,
            )
            if self.truncate_log_q:
                above_min_log_q = log_q > min_log_q
                logger.debug(
                    "Discarding %s samples below log_q_min",
                    self.drawsize - above_min_log_q.sum(),
                )
                x, log_q = get_subset_arrays(above_min_log_q, x, log_q)

            if self.enforce_likelihood_threshold:
                x["logL"] = self.model.batch_evaluate_log_likelihood(
                    x, unit_hypercube=self.map_to_unit_hypercube
                )
                above_threshold = x["logL"] > log_likelihood_threshold
                x, log_q = get_subset_arrays(above_threshold, x, log_q)
                logger.debug(
                    "Accepting %s / %s samples above logL threshold",
                    len(x),
                    self.drawsize,
                )
            # Handle case where all samples have been discarded
            if not len(x):
                logger.warning(
                    "All samples were discard before performing rejection "
                    "sampling."
                )
                continue
            log_w = self.compute_weights(x, log_q)

            if self.accumulate_weights:
                samples = np.concatenate([samples, x])
                log_weights = np.concatenate([log_weights, log_w])
                log_constant = max(np.nanmax(log_w), log_constant)
                log_n_expected = logsumexp(log_weights - log_constant)

                logger.debug(
                    "Drawn %s - n expected: %s / %s",
                    samples.size,
                    np.exp(log_n_expected),
                    n_samples,
                )

                # Only try rejection sampling if we expected to accept enough
                # points. In the case where we don't, we continue drawing
                # samples
                if log_n_expected >= log_n:
                    log_u = np.log(self.rng.random(len(log_weights)))
                    accept = (log_weights - log_constant) > log_u
                    n_accepted = np.sum(accept)
                if n_proposed > max_samples:
                    logger.warning("Reached max samples (%s)", max_samples)
                    break

            else:
                log_w -= log_w.max()
                log_u = np.log(self.rng.random(len(log_w)))
                accept = log_w > log_u
                n_accept_batch = accept.sum()
                m = min(n_samples - n_accepted, n_accept_batch)
                samples[n_accepted : n_accepted + m] = x[accept][:m]
                n_accepted += n_accept_batch
                logger.debug("n accepted: %s / %s", n_accepted, n_samples)

        if self.accumulate_weights:
            if accept is None or len(accept) != len(samples):
                log_u = np.log(self.rng.random(len(log_weights)))
                accept = (log_weights - log_constant) > log_u
            logger.debug("Total number of samples: %s", samples.size)
            n_accepted = np.sum(accept)
            self.x = samples[accept][:n_samples]
        else:
            self.x = samples[:n_samples]

        self.samples = self.convert_to_samples(self.x, plot=plot)

        if self._plot_pool and plot:
            self.plot_pool(self.samples)

        self.population_time += datetime.datetime.now() - st
        logger.debug("Evaluating log-likelihoods")
        if not self.enforce_likelihood_threshold:
            self.samples["logL"] = self.model.batch_evaluate_log_likelihood(
                self.samples
            )
        if self.check_acceptance:
            self.acceptance.append(
                self.compute_acceptance(worst_point["logL"])
            )
            logger.debug(f"Current acceptance {self.acceptance[-1]}")

        self.indices = self.rng.permutation(self.samples.size).tolist()
        self.population_acceptance = n_accepted / n_proposed
        self.populated_count += 1
        self.populated = True
        self._checked_population = False
        logger.debug(f"Proposal populated with {len(self.indices)} samples")
        logger.debug(
            f"Overall proposal acceptance: {self.x.size / n_proposed:.4}"
        )

    def get_alt_distribution(self):
        """
        Get a distribution for the latent prior used to draw samples.
        """
        if self.latent_prior in ["uniform_nsphere", "uniform_nball"]:
            return get_uniform_distribution(
                self.dims, self.r * self.fuzz, device=self.flow.device
            )

    def reset(self):
        """Reset the proposal"""
        super().reset()
        self.r = np.nan
        self.alt_dist = None
        self._populate_dist = None
        self._draw_func = None
        self.alt_dist = None

    def __getstate__(self):
        state = super().__getstate__()
        state["_draw_func"] = None
        state["_populate_dist"] = None
        state["alt_dist"] = None
        return state



# -*- coding: utf-8 -*-
"""Test methods related to initialising and resuming the proposal method"""

from unittest.mock import patch

import numpy as np
import pytest

from nessai.proposal import FlowProposal


def test_init(model):
    """Test init with some kwargs"""
    fp = FlowProposal(model, poolsize=1000)
    assert fp.model == model
    assert fp.poolsize == 1000


@pytest.mark.parametrize("populated", [False, True])
@pytest.mark.parametrize("mask", [None, [1, 0]])
def test_get_state(proposal, populated, mask):
    """Test the get state method used for pickling the proposal.

    Tests cases where the proposal is and isn't populated.
    """
    parent_state = {"a": "val"}
    with patch(
        "nessai.proposal.flowproposal.base.BaseFlowProposal.__getstate__",
        return_value=parent_state,
    ) as mock:
        state = FlowProposal.__getstate__(proposal)

    mock.assert_called_once()
    assert state["_draw_func"] is None
    assert state["_populate_dist"] is None
    assert state["alt_dist"] is None
    assert state["a"] == "val"


@pytest.mark.integration_test
@pytest.mark.parametrize("reparameterisation", [False, True])
@pytest.mark.parametrize("init", [False, True])
@pytest.mark.parametrize(
    "latent_prior", ["truncated_gaussian", "uniform_nball", "gaussian"]
)
def test_resume_pickle(model, tmpdir, reparameterisation, init, latent_prior):
    """Test pickling and resuming the proposal.

    Tests both with and without reparameterisations and before and after
    initialise has been called.
    """
    import pickle

    output = tmpdir.mkdir("test_integration")
    if reparameterisation:
        reparameterisations = {"default": {"parameters": model.names}}
    else:
        reparameterisations = None

    if latent_prior != "truncated_gaussian":
        constant_volume_mode = False
    else:
        constant_volume_mode = True

    proposal = FlowProposal(
        model,
        poolsize=1000,
        plot=False,
        expansion_fraction=1,
        output=output,
        reparameterisations=reparameterisations,
        latent_prior=latent_prior,
        constant_volume_mode=constant_volume_mode,
    )
    if init:
        proposal.initialise()

    proposal.mask = None
    proposal.resume_populated = False

    proposal_data = pickle.dumps(proposal)
    proposal_re = pickle.loads(proposal_data)
    proposal_re.resume(model, {})

    assert proposal._plot_pool == proposal_re._plot_pool
    assert proposal._plot_training == proposal_re._plot_training

    if init:
        assert proposal.fuzz == proposal_re.fuzz
        assert proposal.prime_parameters == proposal_re.rescaled_names


def test_reset(proposal):
    """Test reset method"""
    with patch(
        "nessai.proposal.flowproposal.base.BaseFlowProposal.reset"
    ) as mock:
        FlowProposal.reset(proposal)
    mock.assert_called_once()
    assert proposal.r is np.nan
    assert proposal.alt_dist is None


@pytest.mark.timeout(60)
@pytest.mark.integration_test
@pytest.mark.parametrize(
    "latent_prior", ["truncated_gaussian", "uniform_nball", "gaussian"]
)
def test_reset_integration(tmpdir, model, latent_prior):
    """Test reset method iteration with other methods"""
    flow_config = dict(
        n_neurons=1,
        n_blocks=1,
        n_layers=1,
        batch_norm_between_layers=False,
        linear_transform=None,
    )
    training_config = dict(patience=20)
    output = str(tmpdir.mkdir("reset_integration"))
    poolsize = 2
    drawsize = 100
    if latent_prior != "truncated_gaussian":
        constant_volume_mode = False
    else:
        constant_volume_mode = True
    proposal = FlowProposal(
        model,
        output=output,
        flow_config=flow_config,
        training_config=training_config,
        plot=False,
        poolsize=poolsize,
        drawsize=drawsize,
        expansion_fraction=None,
        latent_prior=latent_prior,
        constant_volume_mode=constant_volume_mode,
    )

    modified_proposal = FlowProposal(
        model,
        output=output,
        flow_config=flow_config,
        training_config=training_config,
        plot=False,
        poolsize=poolsize,
        drawsize=drawsize,
        expansion_fraction=None,
        latent_prior=latent_prior,
        constant_volume_mode=constant_volume_mode,
    )
    proposal.initialise()
    modified_proposal.initialise()

    modified_proposal.populate(model.new_point())
    modified_proposal.reset()

    # attributes that should be different
    ignore = [
        "population_time",
        "_reparameterisation",
        "rng",
    ]

    d1 = proposal.__getstate__()
    d2 = modified_proposal.__getstate__()

    for key in ignore:
        d1.pop(key)
        d2.pop(key)

    assert d1 == d2



# -*- coding: utf-8 -*-
"""Test different properties in BaseFlowProposal"""

import numpy as np
import pytest

from nessai import config
from nessai.proposal.flowproposal.base import BaseFlowProposal

EXTRA_PARAMS_DTYPE = [
    (nsp, d)
    for nsp, d in zip(
        config.livepoints.non_sampling_parameters,
        config.livepoints.non_sampling_dtype,
    )
]


def test_poolsize(proposal):
    """Test poolsize property"""
    proposal._poolsize = 10
    proposal._poolsize_scale = 2
    assert BaseFlowProposal.poolsize.__get__(proposal) == 20


def test_dims(proposal):
    """Test dims property"""
    proposal.parameters = ["x", "y"]
    assert BaseFlowProposal.dims.__get__(proposal) == 2


def test_rescaled_dims(proposal):
    """Test rescaled_dims property"""
    proposal.prime_parameters = ["x", "y"]
    assert BaseFlowProposal.rescaled_dims.__get__(proposal) == 2


def test_dtype(proposal):
    """Test dims property"""
    proposal.parameters = ["x", "y"]
    proposal._x_dtype = None
    assert (
        BaseFlowProposal.x_dtype.__get__(proposal)
        == [("x", "f8"), ("y", "f8")] + EXTRA_PARAMS_DTYPE
    )


def test_prime_dtype(proposal):
    """Test dims property"""
    proposal.prime_parameters = ["x", "y"]
    proposal._x_prime_dtype = None
    assert (
        BaseFlowProposal.x_prime_dtype.__get__(proposal)
        == [("x", "f8"), ("y", "f8")] + EXTRA_PARAMS_DTYPE
    )


def test_population_dtype(proposal):
    """Test dims property"""
    proposal.x_dtype = [
        ("x", "f8"),
        ("y", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]
    proposal.use_x_prime_prior = False
    assert BaseFlowProposal.population_dtype.__get__(proposal) == [
        ("x", "f8"),
        ("y", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]


def test_population_dtype_prime_prior(proposal):
    """Test dims property"""
    proposal.x_prime_dtype = [
        ("x_p", "f8"),
        ("y_p", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]
    proposal.use_x_prime_prior = True
    assert BaseFlowProposal.population_dtype.__get__(proposal) == [
        ("x_p", "f8"),
        ("y_p", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]


@pytest.mark.parametrize(
    "flow_config, expected", [(None, {}), ({"a": 1}, {"a": 1})]
)
def test_flow_config_setter(proposal, flow_config, expected):
    BaseFlowProposal.flow_config.__set__(proposal, flow_config)
    assert proposal._flow_config == expected


def test_training_config(proposal):
    config = {"a": 1}
    proposal._training_config = config
    assert BaseFlowProposal.training_config.__get__(proposal) is config


@pytest.mark.parametrize(
    "training_config, expected", [(None, {}), ({"a": 1}, {"a": 1})]
)
def test_training_config_setter(proposal, training_config, expected):
    BaseFlowProposal.training_config.__set__(proposal, training_config)
    assert proposal._training_config == expected


def test_prior_bounds_no_unit_hypercube(proposal, model):
    proposal._prior_bounds = None
    proposal.map_to_unit_hypercube = False
    proposal.model = model
    out = BaseFlowProposal.prior_bounds.__get__(proposal)
    assert out is model.bounds
    assert proposal._prior_bounds is model.bounds


def test_prior_bounds_unit_hypercube(proposal, model):
    proposal._prior_bounds = None
    proposal.map_to_unit_hypercube = True
    proposal.model = model
    out = BaseFlowProposal.prior_bounds.__get__(proposal)
    assert list(out.keys()) == model.names
    assert all(np.array_equal(v, np.array([0, 1])) for v in out.values())
    assert proposal._prior_bounds is out



# -*- coding: utf-8 -*-
"""
Functions related to creating live points and converting to other common
data-types.
"""

import logging

import numpy as np
from numpy.lib import recfunctions as rfn

from . import config

logger = logging.getLogger(__name__)


def add_extra_parameters_to_live_points(
    parameters,
    default_values=None,
):
    """Add extra parameters to the live points dtype.

    Extra parameters will be included in the live points dtype that is used
    for constructing/converting to/from live points.

    Parameters
    ----------
    parameters: list
        List of parameters to add.
    default_values: Optional[Union[List, Tuple]]
        List of default values for each parameters. If not specified, default
        values will be set to based on :code: `DEFAULT_FLOAT_VALUE` in
        :code:`nessai.config`.
    """
    if default_values is None:
        default_values = len(parameters) * (
            config.livepoints.default_float_value,
        )
    else:
        default_values = tuple(default_values)
    for p, dv in zip(parameters, default_values):
        if p not in config.livepoints.extra_parameters:
            config.livepoints.extra_parameters.append(p)
            config.livepoints.extra_parameters_defaults = (
                config.livepoints.extra_parameters_defaults + (dv,)
            )
            config.livepoints.extra_parameters_dtype.append(
                config.livepoints.default_float_dtype
            )
        else:
            logger.warning(
                f"Extra parameter `{p}` has already been added. Skipping."
                "Call `reset_extra_live_points_parameters` to reset the values"
                " and add this parameter."
            )

    logger.debug(
        "Updated non-sampling parameters: "
        f"{config.livepoints.non_sampling_parameters}"
    )
    logger.debug(
        "Updated defaults for non-sampling parameters: "
        f"{config.livepoints.non_sampling_defaults}"
    )
    config.livepoints.reset_properties()


def reset_extra_live_points_parameters():
    """Reset the extra live points parameters."""
    logger.debug("Resetting extra parameters")
    config.livepoints.reset()


def get_dtype(names, array_dtype=None, non_sampling_parameters=True):
    """
    Get a list of tuples containing the dtypes for the structured array

    Parameters
    ----------
    names : list of str
        Names of parameters
    array_dtype : Optional[str]
        dtype to use
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    numpy.dtype
        A instance of :code:`numpy.dtype`.
    """
    if array_dtype is None:
        array_dtype = config.livepoints.default_float_dtype
    dtype = [(n, array_dtype) for n in names]
    if non_sampling_parameters:
        dtype += list(
            zip(
                config.livepoints.non_sampling_parameters,
                config.livepoints.non_sampling_dtype,
            )
        )
    return np.dtype(dtype)


def empty_structured_array(
    n, names=None, dtype=None, non_sampling_parameters=True
):
    """Get an empty structured array with the extra parameters initialised.

    Parameters
    ----------
    n : int
        Length of the structured array
    dtype : Optional[list]
        Dtype to use. Must contain the non-sampling parameters.
    names : Optional[list]
        Names of fields (excluding non-sampling parameters) to construct the
        dtype. Must be specified if :code:`dtype` is not specified.
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    np.ndarray
        Structured array with the all parameters initialised to their
        default values.
    """
    if dtype is None:
        dtype = get_dtype(
            names, non_sampling_parameters=non_sampling_parameters
        )
    else:
        dtype = np.dtype(dtype)
        names = [
            nm
            for nm in dtype.names
            if nm not in config.livepoints.non_sampling_parameters
        ]
    struct_array = np.empty((n), dtype=dtype)
    if n == 0:
        return struct_array
    struct_array[names] = config.livepoints.default_float_value
    if non_sampling_parameters:
        try:
            for nm, v in zip(
                config.livepoints.non_sampling_parameters,
                config.livepoints.non_sampling_defaults,
            ):
                struct_array[nm] = v
        except ValueError:
            raise ValueError(
                "Could not create empty structured array. Maybe the "
                "non-sampling parameters are missing?"
            )
    return struct_array


def live_points_to_array(live_points, names=None, copy=False):
    """
    Converts live points to unstructured arrays for training.

    Parameters
    ----------
    live_points : structured_array
        Structured array of live points
    names : list of str or None
        If None all fields in the structured array are added to the dictionary
        else only those included in the list are added.
    copy : bool
        If true, returns a copy. If false, returns a view. See numpy
        documentation for
        :code:`numpy.lib.recfunctions.structured_to_unstructured` for more
        details.

    Returns
    -------
    np.ndarray
        Unstructured numpy array
    """
    if names is None:
        names = list(live_points.dtype.names)
    return rfn.structured_to_unstructured(live_points[names], copy=copy)


def parameters_to_live_point(parameters, names, non_sampling_parameters=True):
    """
    Take a list or array of parameters for a single live point
    and converts them to a live point.

    Returns an empty array with the correct fields if len(parameters) is zero

    Parameters
    ----------
    parameters : tuple
        Float point values for each parameter
    names : tuple
        Names for each parameter as strings
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by names plus logP and logL
    """
    if not len(parameters):
        return empty_structured_array(
            0, names, non_sampling_parameters=non_sampling_parameters
        )
    else:
        if non_sampling_parameters:
            parameters = [
                (*parameters, *config.livepoints.non_sampling_defaults)
            ]
        else:
            parameters = [tuple(parameters)]
        return np.array(
            parameters,
            dtype=get_dtype(
                names,
                config.livepoints.default_float_dtype,
                non_sampling_parameters=non_sampling_parameters,
            ),
        )


def numpy_array_to_live_points(array, names, non_sampling_parameters=True):
    """
    Convert a numpy array to a numpy structure array with the correct fields

    Parameters
    ----------
    array : np.ndarray
        Instance of np.ndarray to convert to a structured array
    names : tuple
        Names for each parameter as strings
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by names plus logP and logL
    """
    if array.size == 0:
        return empty_structured_array(
            0, names=names, non_sampling_parameters=non_sampling_parameters
        )
    if array.ndim == 1:
        array = array[np.newaxis, :]
    struct_array = empty_structured_array(
        len(array),
        names=names,
        non_sampling_parameters=non_sampling_parameters,
    )
    for i, n in enumerate(names):
        struct_array[n] = array[..., i]
    return struct_array


def dict_to_live_points(d, non_sampling_parameters=True):
    """Convert a dictionary with parameters names as keys to live points.

    Assumes all entries have the same length. Also, determines if points are
    array-like by checking if the first value has `__len__` attribute and
    the number of dimensions is not zero, if not the dictionary is assumed to
    contain a single live point.

    Parameters
    ----------
    d : dict
        Dictionary with parameters names as keys and values that correspond
        to one or more parameters
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by names plus logP and logL
    """
    values = tuple(d.values())
    is_array_like = (
        hasattr(values[0], "__len__") and not np.ndim(values[0]) == 0
    )
    if not is_array_like:
        if non_sampling_parameters:
            values = (*values, *config.livepoints.non_sampling_defaults)
        return np.array(
            [values],
            dtype=get_dtype(
                d.keys(),
                config.livepoints.default_float_dtype,
                non_sampling_parameters=non_sampling_parameters,
            ),
        )
    else:
        n = len(values[0])
        array = empty_structured_array(
            n,
            names=list(d.keys()),
            non_sampling_parameters=non_sampling_parameters,
        )
        for k, v in d.items():
            array[k] = v
        return array


def live_points_to_dict(live_points, names=None):
    """
    Convert a structured array of live points to a dictionary with
    a key per field.

    Parameters
    ----------
    live_points : structured_array
        Array of live points
    names : list of str or None
        If None all fields in the structured array are added to the dictionary
        else only those included in the list are added.

    Returns
    -------
    dict
        Dictionary of live points
    """
    if names is None:
        names = live_points.dtype.names
    return {f: live_points[f] for f in names}


def dataframe_to_live_points(df, non_sampling_parameters=True):
    """Convert and pandas dataframe to live points.

    Adds the non-sampling parameter initialised to their defaults.

    Based on this answer on Stack Exchange:
    https://stackoverflow.com/a/51280608

    Parameters
    ----------
    df : :obj:`pandas.DataFrame`
        Pandas DataFrame to convert to live points
    non_sampling_parameters : bool
        Indicates whether non-sampling parameters should be included.

    Returns
    -------
    structured_array
        Numpy structured array with fields given by column names plus logP and
        logL.
    """
    if non_sampling_parameters:
        extra = config.livepoints.non_sampling_defaults
    else:
        extra = tuple()
    return np.array(
        [tuple(x) + extra for x in df.values],
        dtype=get_dtype(
            list(df.dtypes.index),
            non_sampling_parameters=non_sampling_parameters,
        ),
    )


def _unstructured_view_dtype(x, names):
    """Get the dtype for an unstructured view.

    Parameters
    ----------
    x : numpy.ndarray
        Array on which to base the dtype.
    names : Iterable
        Names for the fields to include in the dtype.

    Returns
    -------
    numpy.dtype
        Dtype containing the specified fields.
    """
    return np.dtype({name: x.dtype.fields[name] for name in names})


def unstructured_view(x, names=None, dtype=None):
    """Get an unstructured view of a live points containing certain parameters.

    This is quicker than converting to a unstructured array and does not
    create a copy of the array.

    Parameters
    ----------
    x : numpy.ndarray
        Structured array.
    names : Optional[Iterable]
        Iterable of parameters to include in the view. Must be specified if
        dtype is None.
    dtype : Optional[numpy.dtype]
        Dtype for constructing the unstructured view.

    Returns
    -------
    numpy.ndarray
        View of x as an unstructured array that contains only the
        parameters in names/dtype. Shape is (x.size, # parameters).
    """
    if dtype is None:
        dtype = _unstructured_view_dtype(x, names)
    return np.ndarray(x.shape, dtype, x, 0, x.strides).view(
        (config.livepoints.default_float_dtype, len(dtype))
    )



from unittest.mock import create_autospec

import numpy as np
import pytest

from nessai.livepoint import dict_to_live_points, empty_structured_array
from nessai.reparameterisations.discrete import Dequantise
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture
def reparam(rng):
    return create_autospec(Dequantise, rng=rng)


def test_dequantise_init():
    reparam = Dequantise(parameters=["a"], prior_bounds={"a": [0, 10]})
    assert reparam.has_pre_rescaling is True
    assert reparam.has_prime_prior is False

    # Bounds should be one more than the max
    assert reparam.bounds["a"][0] == 0
    assert reparam.bounds["a"][1] == 11


def test_pre_rescaling(reparam):
    x = np.array([0, 0, 0, 1, 1, 1])
    out, log_j = Dequantise.pre_rescaling(reparam, x)
    assert all((out[:3] > 0) & (out[:3] < 1))
    assert all((out[3:] > 1) & (out[3:] < 2))
    np.testing.assert_equal(log_j, 0.0)


def test_pre_rescaling_inv(reparam):
    x = np.array([0.5, 1.1, 2.3])
    out, log_j = Dequantise.pre_rescaling_inv(reparam, x)
    np.testing.assert_array_equal(out, np.array([0, 1, 2]))
    np.testing.assert_equal(log_j, 0.0)


def test_dequantise_integration():
    values = [0, 1, 2, 4, 8]
    x = dict_to_live_points({"a": np.random.choice(values, size=10)})

    reparam = Dequantise(
        parameters="a",
        prior_bounds={"a": [0, 8]},
        rescale_bounds=[0, 1],
    )

    x_prime = empty_structured_array(len(x), names=reparam.prime_parameters)
    x_out, x_prime, log_j_out = reparam.reparameterise(
        x, x_prime, np.zeros(len(x))
    )
    assert_structured_arrays_equal(x_out, x)

    assert not np.array_equal(x_prime["a_prime"], x["a"])
    assert all((x_prime["a_prime"] > 0) & (x_prime["a_prime"] < 1))
    # Should have 10 unique values
    assert len(np.unique(x_prime["a_prime"]) == len(x))

    x_re, x_prime_re, log_j_re = reparam.inverse_reparameterise(
        x_out.copy(), x_prime.copy(), np.zeros(len(x))
    )
    assert_structured_arrays_equal(x_prime_re, x_prime)

    assert_structured_arrays_equal(x_re, x)
    np.testing.assert_array_equal(-log_j_re, log_j_out)



"""
Test functions related to training and using the flow.
"""

from unittest.mock import MagicMock

import numpy as np
import pytest

from nessai.proposal import FlowProposal


@pytest.mark.parametrize("log_p", [np.ones(2), np.array([-1, np.inf])])
@pytest.mark.parametrize("discard_nans", [False, True])
@pytest.mark.parametrize("return_z", [False, True])
def test_backward_pass(
    proposal,
    model,
    map_to_unit_hypercube,
    log_p,
    discard_nans,
    return_z,
):
    """Test the forward pass method"""
    n = 2
    if discard_nans:
        acc = int(np.isfinite(log_p).sum())
    else:
        acc = len(log_p)
    x = np.random.randn(n, model.dims)
    z = np.random.randn(n, model.dims)

    def inverse_rescale(a, return_unit_hypercube):
        return a, np.zeros(a.size)

    proposal.inverse_rescale = MagicMock(side_effect=inverse_rescale)
    proposal.prime_parameters = model.names
    proposal.alt_dist = None
    proposal.check_prior_bounds = MagicMock(
        side_effect=lambda a, b, c: (a, b, c)
    )
    proposal.flow = MagicMock()
    proposal.flow.sample_and_log_prob = MagicMock(return_value=[x, log_p])

    out = FlowProposal.backward_pass(
        proposal,
        z,
        discard_nans=discard_nans,
        return_unit_hypercube=map_to_unit_hypercube,
        return_z=return_z,
    )

    x_out = out[0]
    log_p_out = out[1]

    assert len(x_out) == acc
    assert len(log_p_out) == acc
    if return_z:
        assert len(out) == 3
        assert out[2].shape == (acc, model.dims)
    else:
        assert len(out) == 2
    proposal.inverse_rescale.assert_called_once()
    assert (
        proposal.inverse_rescale.call_args.kwargs["return_unit_hypercube"]
        is map_to_unit_hypercube
    )
    proposal.flow.sample_and_log_prob.assert_called_once_with(
        z=z, alt_dist=None
    )


@pytest.mark.parametrize("return_z", [False, True])
def test_backwards_pass_assertion_error(proposal, caplog, return_z):
    proposal.alt_dist = None
    proposal.flow = MagicMock()

    def func(*args, **kwargs):
        raise AssertionError("Domain")

    proposal.flow.sample_and_log_prob = MagicMock(side_effect=func)
    out = FlowProposal.backward_pass(
        proposal, np.random.randn(10, 2), return_z=return_z
    )
    assert all([len(a) == 0 for a in out])
    if return_z:
        assert len(out) == 3
    else:
        assert len(out) == 2
    assert "Domain" in caplog.text



import pytest

from nessai.stopping_criteria import (
    CriterionGroup,
    StoppingCriterion,
    StoppingCriterionRegistry,
)


@pytest.mark.parametrize(
    "comparison, value, tolerance, expected",
    [
        ("<", 4, 5, True),
        ("<=", 5, 5, True),
        (">", 6, 5, True),
        (">=", 5, 5, True),
        ("==", 5, 5, True),
        ("!=", 4, 5, True),
        ("<", 6, 5, False),
        ("<=", 6, 5, False),
        (">", 5, 5, False),
        (">=", 4, 5, False),
        ("==", 4, 5, False),
        ("!=", 5, 5, False),
    ],
)
def test_stopping_criterion_is_met(comparison, value, tolerance, expected):
    crit = StoppingCriterion("test", tolerance, comparison)
    assert crit.is_met(value) == expected


def test_invalid_comparison_operator_raises():
    with pytest.raises(ValueError, match="Invalid comparison operator"):
        StoppingCriterion("bad", 1.0, "===")


def test_criterion_group_and_logic():
    c1 = StoppingCriterion("c1", 5, "<=")
    c2 = StoppingCriterion("c2", 10, ">")
    group = CriterionGroup([c1, c2], mode="and")
    values = {"c1": 5, "c2": 11}
    assert group.is_met(values)


def test_criterion_group_or_logic():
    c1 = StoppingCriterion("c1", 5, "<=")
    c2 = StoppingCriterion("c2", 10, ">")
    group = CriterionGroup([c1, c2], mode="or")
    values = {"c1": 6, "c2": 11}
    assert group.is_met(values)


def test_group_names_and_tolerances():
    c1 = StoppingCriterion("c1", 0.5, "<")
    c2 = StoppingCriterion("c2", 1.5, ">")
    group = CriterionGroup([c1, c2])
    assert group.names == ["c1", "c2"]
    assert group.tolerances == {"c1": 0.5, "c2": 1.5}


def test_group_composition_and_or():
    c1 = StoppingCriterion("a", 1.0, "<")
    c2 = StoppingCriterion("b", 2.0, "!=")
    g1 = CriterionGroup([c1])
    g2 = CriterionGroup([c2])

    combined_and = g1 & g2
    assert isinstance(combined_and, CriterionGroup)
    assert len(combined_and.criteria) == 2
    assert combined_and.mode == "and"

    combined_or = g1 | g2
    assert isinstance(combined_or, CriterionGroup)
    assert len(combined_or.criteria)


def test_composition_and_or():
    c1 = StoppingCriterion("a", 1.0, "<")
    c2 = StoppingCriterion("b", 2.0, "!=")

    combined = c1 & c2
    assert isinstance(combined, CriterionGroup)
    assert len(combined.criteria) == 2
    assert combined.mode == "and"

    combined = c1 | c2
    assert isinstance(combined, CriterionGroup)
    assert len(combined.criteria) == 2
    assert combined.mode == "or"


class DummyCriterion(StoppingCriterion):
    def __init__(self, name="dummy", tolerance=1.0, comparison="<"):
        super().__init__(name, tolerance, comparison)


@pytest.mark.parametrize(
    "names",
    [
        ("single_name",),
        ("name_one", "name_two"),
        ("MixedCase", "lowercase", "UPPERCASE"),
    ],
)
def test_register_and_get_criterion(names):
    # Register dummy criterion under multiple names
    StoppingCriterionRegistry.register(*names)(DummyCriterion)

    for name in names:
        retrieved = StoppingCriterionRegistry.get(name)
        assert isinstance(retrieved, DummyCriterion)
        assert retrieved.name == "dummy"


@pytest.mark.parametrize(
    "lookup_name, expected_cls",
    [
        ("ess", "ESS"),
        ("log_evidence_ratio", "LogEvidenceRatio"),
        ("log_evidence_ratio_nested_samples", "LogEvidenceRatioNestedSamples"),
        ("ratio", "LogEvidenceRatio"),
        ("Z_err", "EvidenceError"),
        ("difference_log_evidence", "DifferenceLogEvidence"),
        ("fractional_error", "FractionalError"),
    ],
)
def test_get_registered_criteria_classes(lookup_name, expected_cls):
    retrieved = StoppingCriterionRegistry.get(lookup_name)
    assert retrieved.__class__.__name__ == expected_cls


def test_registry_case_insensitivity():
    StoppingCriterionRegistry.register("CaseTest")(DummyCriterion)
    assert isinstance(
        StoppingCriterionRegistry.get("casetest"), DummyCriterion
    )
    assert isinstance(
        StoppingCriterionRegistry.get("CASETEST"), DummyCriterion
    )
    assert isinstance(
        StoppingCriterionRegistry.get("CaseTest"), DummyCriterion
    )


def test_list_available_includes_defaults():
    available = StoppingCriterionRegistry.list_available()
    # These are defaults from your submodule
    expected_keys = {
        "ess",
        "log_evidence_ratio",
        "ratio",
        "log_evidence_ratio_nested_samples",
        "ratio_ns",
        "z_err",
        "evidence_error",
        "dlogz",
        "difference_log_evidence",
        "fractional_error",
    }
    assert expected_keys.issubset(set(available))


def test_get_unregistered_name_raises():
    with pytest.raises(
        ValueError, match="No registered criterion with name 'unknown_name'"
    ):
        StoppingCriterionRegistry.get("unknown_name")



"""Test methods related to initialising and resuming the proposal method"""

import os
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from nessai.proposal.flowproposal.base import BaseFlowProposal


@pytest.mark.parametrize(
    "value, expected", [(True, True), (False, False), (None, False)]
)
def test_init_use_default_reparams(model, proposal, value, expected):
    """Assert use_default_reparameterisations is set correctly"""
    proposal.use_default_reparameterisations = False
    BaseFlowProposal.__init__(
        proposal, model, poolsize=10, use_default_reparameterisations=value
    )
    assert proposal.use_default_reparameterisations is expected


def test_initialise(tmpdir, proposal):
    """Test the initialise method"""
    p = tmpdir.mkdir("test")
    proposal.initialised = False
    proposal.output = os.path.join(p, "output")
    proposal.flow_config = {}
    proposal.training_config = {}
    proposal.set_rescaling = MagicMock()
    proposal.verify_rescaling = MagicMock()
    proposal.update_flow_config = MagicMock()
    fm = MagicMock()
    fm.initialise = MagicMock()
    proposal._FlowModelClass = MagicMock(new=fm)

    BaseFlowProposal.initialise(proposal, resumed=False)

    proposal.set_rescaling.assert_called_once()
    proposal.verify_rescaling.assert_called_once()
    proposal.update_flow_config.assert_called_once()
    proposal._FlowModelClass.assert_called_once_with(
        flow_config=proposal.flow_config,
        training_config=proposal.training_config,
        output=proposal.output,
        rng=proposal.rng,
    )
    proposal.flow.initialise.assert_called_once()
    assert proposal.populated is False
    assert proposal.initialised
    assert os.path.exists(os.path.join(p, "output"))


def test_resume(proposal):
    """Test the resume method."""
    proposal.initialise = MagicMock()
    proposal.mask = [1, 0]
    proposal.update_bounds = False
    proposal.weights_file = None
    model = MagicMock()
    with patch("nessai.proposal.base.Proposal.resume") as mock:
        BaseFlowProposal.resume(proposal, model, {})
    mock.assert_called_once_with(model)
    proposal.initialise.assert_called_once()
    assert np.array_equal(proposal.flow_config["mask"], np.array([1, 0]))


@patch("os.path.exists", return_value=True)
def test_resume_w_weights(osexist, proposal):
    """Test the resume method with weights"""
    proposal.initialise = MagicMock()
    proposal.flow = MagicMock()
    proposal.mask = None
    proposal.update_bounds = False
    proposal.weights_file = None
    model = MagicMock()
    with patch("nessai.proposal.base.Proposal.resume") as mock:
        BaseFlowProposal.resume(proposal, model, {}, weights_file="weights.pt")
    mock.assert_called_once_with(model)
    osexist.assert_called_once_with("weights.pt")
    proposal.initialise.assert_called_once()
    proposal.flow.reload_weights.assert_called_once_with("weights.pt")


@pytest.mark.parametrize("populated", [False, True])
@pytest.mark.parametrize("mask", [None, [1, 0]])
def test_get_state(proposal, populated, mask):
    """Test the get state method used for pickling the proposal.

    Tests cases where the proposal is and isn't populated.
    """

    proposal.populated = populated
    proposal.indices = [1, 2]
    proposal._reparameterisation = MagicMock()
    proposal.model = MagicMock()
    proposal._flow_config = {}
    proposal.initialised = True
    proposal.flow = MagicMock()
    proposal.flow.weights_file = "file"

    if mask is not None:
        proposal.flow.flow_config = {"mask": mask}

    state = BaseFlowProposal.__getstate__(proposal)

    assert state["resume_populated"] is populated
    assert state["initialised"] is False
    assert state["weights_file"] == "file"
    assert state["mask"] is mask
    assert "model" not in state
    assert "flow" not in state
    assert "_flow_config" not in state


def test_reset(proposal):
    """Test reset method"""
    proposal.x = 1
    proposal.samples = 2
    proposal.populated = True
    proposal.populated_count = 10
    proposal._reparameterisation = MagicMock()
    BaseFlowProposal.reset(proposal)
    assert proposal.x is None
    assert proposal.samples is None
    assert proposal.populated is False
    assert proposal.populated_count == 0
    assert proposal._checked_population
    proposal._reparameterisation.reset.assert_called_once()


def test_populate_error(proposal):
    with pytest.raises(NotImplementedError):
        BaseFlowProposal.populate(proposal, 1.0, n_samples=10)



from unittest.mock import create_autospec

import pytest

from nessai.samplers.nestedsampler import NestedSampler


@pytest.fixture(scope="function")
def sampler(model, rng):
    s = create_autospec(NestedSampler)
    s.nlive = 100
    s.model = model
    s.store_live_points = False
    s.rng = rng
    return s



# -*- coding: utf-8 -*-
"""
Object and functions to handle training the normalising flow.
"""

import copy
import logging
import os
import shutil
from typing import Optional, Tuple, Union

import numpy as np
import torch
from torch.nn.utils import clip_grad_norm_

from ..flows import configure_model, reset_permutations, reset_weights
from ..flows.base import BaseFlow
from ..plot import plot_loss
from ..utils import compute_minimum_distances, save_to_json
from .utils import update_config

logger = logging.getLogger(__name__)


class FlowModel:
    """
    Object that contains the normalising flows and handles training and data
    pre-processing.

    Does NOT use structured arrays for live points, \
            :obj:`~nessai.proposal.base.Proposal`
    object should act as the interface between structured used by the sampler
    and unstructured arrays of live points used for training.

    Parameters
    ----------
    config : dict, optional
        Configuration used for the normalising flow. If None, default values
        are used.
    output : str, optional
        Path for output, this includes weights files and the loss plot. If
        not specified, the current working directory is used.
    """

    noise_scale = None
    noise_type = None
    model: BaseFlow = None

    def __init__(
        self,
        flow_config: Union[dict, None] = None,
        training_config: Union[dict, None] = None,
        output: Union[str, None] = None,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        if output is None:
            output = os.getcwd()
        self.model = None
        if rng is None:
            logger.debug("No rng specified, using the default rng.")
            rng = np.random.default_rng()
        self.rng = rng
        self.initialised = False
        self.output = output
        os.makedirs(self.output, exist_ok=True)
        self.setup_from_input_dict(
            flow_config=flow_config,
            training_config=training_config,
        )
        self.weights_file = None
        self._batch_size = None
        self._optimiser = None

    def setup_from_input_dict(self, flow_config, training_config):
        """
        Setup the trainer from a dictionary, all keys in the dictionary are
        added as methods to the object. Input is automatically saved.

        Parameters
        ----------
        config : dict
            Dictionary with parameters that are used to update the defaults.
        """
        self.flow_config, self.training_config = update_config(
            flow_config=flow_config, training_config=training_config
        )
        logger.debug(f"Flow configuration: {self.flow_config}")
        logger.debug(f"Training configuration: {self.training_config}")

        save_to_json(
            self.flow_config, os.path.join(self.output, "flow_config.json")
        )
        save_to_json(
            self.training_config,
            os.path.join(self.output, "training_config.json"),
        )

    def update_mask(self):
        """Method to update the ask upon calling ``initialise``

        By default the mask is left unchanged.
        """
        pass

    def get_optimiser(self, optimiser=None, **kwargs):
        """
        Get the optimiser and ensure it is always correctly initialised.

        By default kwargs will include any optimiser specific values,
        the learning rate and any values included in :code:`optimiser_kwargs`.
        These will then be updated with any additional keyword arguments that
        are passed.

        Returns
        -------
        :obj:`torch.optim.Adam`
            Instance of the Adam optimiser from torch.optim
        """
        optimisers = {
            "adam": (torch.optim.Adam, {"weight_decay": 1e-6}),
            "adamw": (torch.optim.AdamW, {}),
            "sgd": (torch.optim.SGD, {}),
        }
        if self.model is None:
            raise RuntimeError("Cannot initialise optimiser before model")
        if optimiser is None:
            optimiser = self.optimiser
        optim, default_kwargs = optimisers.get(optimiser.lower())
        default_kwargs["lr"] = self.training_config["lr"]
        default_kwargs.update(self.optimiser_kwargs)
        default_kwargs.update(kwargs)
        return optim(
            self.model.parameters(),
            **default_kwargs,
        )

    @property
    def optimiser_kwargs(self) -> dict:
        kwds = self.training_config.get("optimiser_kwargs")
        if kwds is None:
            kwds = {}
        return kwds

    @property
    def optimiser(self) -> str:
        return self.training_config["optimiser"]

    def initialise(self):
        """
        Initialise the model and optimiser.

        This includes:

            - Updating the model configuration
            - Initialising the normalising flow
            - Initialising the optimiser
            - Configuring the inference device
        """
        self.update_mask()
        self.model = configure_model(self.flow_config)
        logger.debug("Flow model:")
        logger.debug(self.model)
        self.device = torch.device(
            self.training_config.get("device_tag", "cpu")
        )
        # Set the default location for the model
        self.model.device = self.device
        logger.debug(f"Training device: {self.device}")
        self.inference_device = torch.device(
            self.flow_config.get("inference_device_tag", self.device)
            or self.device
        )
        logger.debug(f"Inference device: {self.inference_device}")

        self._optimiser = self.get_optimiser()
        self.initialised = True

    def move_to(self, device, update_default=False):
        """Move the flow to a different device.

        Parameters
        ----------
        device : str
            Device to move flow to.
        update_default : bool, optional
            If True, the default device for the flow (and data) will be
            updated.
        """
        device = torch.device(device)
        self.model.to(device)
        if update_default:
            self.device = device

    @staticmethod
    def check_batch_size(x, batch_size, min_fraction=0.1):
        """Check that the batch size is valid.

        Tries to ensure that the last batch is at least a minimum fraction of
        the size of the batch size.

        Parameters
        ----------
        x : numpy.ndarray
            Training data.
        batch_size : int
            The user-specified batch size
        min_fraction : float
            Fraction of user-specified batch size to use a lower limit.
        """
        logger.debug("Checking batch size")
        if batch_size == 1:
            raise ValueError("Cannot use a batch size of 1!")
        min_batch_size = int(min_fraction * batch_size)
        final_batch_size = len(x) % batch_size
        if final_batch_size and (final_batch_size < min_batch_size):
            logger.debug(
                "Adjusting batch size to ensure final batch has at least "
                f"{min_batch_size} samples in it."
            )
            while True:
                batch_size -= 1
                final_batch_size = len(x) % batch_size
                if batch_size < 2:
                    raise RuntimeError("Could not find a valid batch size")
                elif (final_batch_size == 0) or (
                    final_batch_size >= min_batch_size
                ):
                    break
                elif (batch_size <= min_batch_size) and final_batch_size > 1:
                    logger.warning(
                        f"Batch size is less than {min_batch_size} but valid. "
                        f"Setting batch size to: {batch_size}"
                    )
                    break
        logger.debug(f"Using valid batch size of: {batch_size}")
        return batch_size

    def prep_data(
        self,
        samples,
        val_size,
        batch_size,
        weights=None,
        use_dataloader=False,
        conditional=None,
    ):
        """
        Prep data and return dataloaders for training

        Parameters
        ----------
        samples : array_like
            Array of samples to split in to training and validation.
        val_size : float
            Float between 0 and 1 that defines the fraction of data used for
            validation.
        batch_size : int
            Batch size used when constructing dataloaders.
        use_dataloader : bool, optional
            If True data is returned in a dataloader else a tensor is returned.
        weights : array_like, optional
            Array of weights for each sample, weights will used when computing
            the loss.

        Returns
        -------
        train_data, val_data :
            Training and validation data as either a tensor or dataloader
        """
        if not self.initialised:
            self.initialise()

        if not np.isfinite(samples).all():
            raise ValueError("Cannot train with non-finite samples!")

        if weights is not None and conditional is not None:
            raise RuntimeError("weights and conditional inputs not supported")

        idx = self.rng.permutation(samples.shape[0])
        samples = samples[idx]
        if weights is not None:
            if not np.isfinite(weights).all():
                raise ValueError("Weights contain non-finite values!")
            weights = weights[idx]
            use_dataloader = True
        if conditional is not None:
            conditional = conditional[idx]
            use_dataloader = True

        logger.debug("N input samples: {}".format(len(samples)))

        # setup data loading
        if val_size is None:
            val_size = 0
        n = int((1 - val_size) * samples.shape[0])
        x_train, x_val = samples[:n], samples[n:]
        if weights is not None:
            weights_train = weights[:n]
            weights_val = weights[n:]
        if conditional is not None:
            conditional_train = conditional[:n]
            conditional_val = conditional[n:]
        logger.debug(f"{x_train.shape} training samples")
        logger.debug(f"{x_val.shape} validation samples")

        if isinstance(batch_size, bool) or not isinstance(batch_size, int):
            if batch_size == "all" or batch_size is None:
                batch_size = x_train.shape[0]
            else:
                raise RuntimeError(f"Unknown batch size: {batch_size}")

        batch_size = self.check_batch_size(x_train, batch_size)
        self._batch_size = batch_size
        # Validation size cannot be bigger than the batch size
        val_batch_size = min(len(x_val), batch_size) if len(x_val) else None

        dtype = torch.get_default_dtype()
        logger.debug(f"Using dtype {dtype} for tensors")
        if use_dataloader:
            logger.debug("Using dataloaders")
            train_tensors = [torch.from_numpy(x_train).type(dtype)]
            val_tensors = [torch.from_numpy(x_val).type(dtype)]

            if weights is not None:
                train_tensors.append(
                    torch.from_numpy(weights_train).type(dtype)
                )
                val_tensors.append(torch.from_numpy(weights_val).type(dtype))

            if conditional is not None:
                train_tensors.append(
                    torch.from_numpy(conditional_train).type(dtype)
                )
                val_tensors.append(
                    torch.from_numpy(conditional_val).type(dtype)
                )

            train_dataset = torch.utils.data.TensorDataset(*train_tensors)
            train_data = torch.utils.data.DataLoader(
                train_dataset, batch_size=batch_size, shuffle=True
            )

            val_dataset = torch.utils.data.TensorDataset(*val_tensors)
            val_data = torch.utils.data.DataLoader(
                val_dataset, batch_size=val_batch_size, shuffle=False
            )
        else:
            logger.debug("Using tensors")
            train_data = torch.from_numpy(x_train).type(dtype).to(self.device)
            val_data = torch.from_numpy(x_val).type(dtype).to(self.device)

        return train_data, val_data, batch_size

    def end_iteration(self):
        """Calls any functions that should be applied at the end of the \
            iteration.

        This functions is called after the flow has been updated on all batches
        of data but before the validation step.

        Calls :py:meth:`nessai.flows.base.BaseFlow.end_iteration`
        """
        self.model.end_iteration()

    def _train(
        self,
        train_data,
        noise_scale=0.0,
        is_dataloader=False,
        weighted=False,
        is_conditional=False,
    ):
        """
        Loop over the data and update the weights

        Parameters
        ----------
        train_data : :obj:`torch.util.data.Dataloader` or :obj:`torch.Tensor`
            Training data. If a tensor is provided, it is split into batches
            using the batch size.
        noise_scale : float, optional
            Scale of Gaussian noise added to data.
        is_dataloader : bool, optional
            Must be True when using a dataloader
        weighted : bool
            If True the weighted KL will be used to compute the loss. Requires
            data to include weights.
        is_conditional : bool
            If True, the conditional argument will be passed to the flow.

        Returns
        -------
        float
            Mean of training loss for each batch.
        """
        model = self.model
        model.train()
        train_loss = 0

        if hasattr(model, "loss_function"):
            loss_fn = model.loss_function
        elif weighted:

            def loss_fn(data, weights):
                return -torch.sum(model.log_prob(data) * weights) / torch.sum(
                    weights
                )

        else:

            def loss_fn(data, conditional):
                return -model.log_prob(data, conditional).mean()

        if not is_dataloader:
            p = torch.randperm(train_data.shape[0])
            train_data = train_data[p, :].split(self._batch_size)

        n = 0
        for data in train_data:
            if is_dataloader:
                x = data[0].to(self.device)
            else:
                x = data
            if noise_scale:
                x = x + noise_scale * torch.randn_like(x)
            for param in model.parameters():
                param.grad = None
            if weighted:
                weights = data[1].to(self.device)
                loss = loss_fn(x, weights)
            else:
                if is_conditional:
                    conditional = data[1].to(self.device)
                else:
                    conditional = None
                loss = loss_fn(x, conditional)
            train_loss += loss.item()
            loss.backward()
            if self.training_config["clip_grad_norm"]:
                clip_grad_norm_(
                    model.parameters(),
                    self.training_config["clip_grad_norm"],
                )
            self._optimiser.step()
            n += 1

        self.end_iteration()

        if self.training_config["annealing"]:
            self.scheduler.step()

        return train_loss / n

    def _validate(
        self,
        val_data,
        is_dataloader=False,
        weighted=False,
        is_conditional=False,
    ):
        """
        Loop over the data and get validation loss

        Parameters
        ----------
        val_data : :obj:`torch.util.data.Dataloader` or :obj:`torch.Tensor
            Dataloader with data to validate on
        is_dataloader : bool, optional
            Boolean to indicate if the data is a dataloader
        weighted : bool
            If True the weighted KL will be used to compute the loss. Requires
            data to include weights.
        is_conditional : bool
            If True, the conditional argument will be passed to the flow.

        Returns
        -------
        float
            Mean of training loss for each batch.
        """
        if not len(val_data):
            return np.nan
        model = self.model
        model.eval()
        val_loss = 0

        if hasattr(model, "loss_function"):
            loss_fn = model.loss_function
        elif weighted:

            def loss_fn(data, weights):
                return -torch.sum(model.log_prob(data) * weights) / torch.sum(
                    weights
                )

        else:

            def loss_fn(data, conditional):
                return -model.log_prob(data, conditional).mean()

        if is_dataloader:
            n = 0
            for data in val_data:
                if is_dataloader:
                    x = data[0].to(self.device)
                if weighted:
                    weights = data[1].to(self.device)
                    with torch.inference_mode():
                        val_loss += loss_fn(x, weights).item()
                else:
                    if is_conditional:
                        conditional = data[1].to(self.device)
                    else:
                        conditional = None
                    with torch.inference_mode():
                        val_loss += loss_fn(x, conditional).item()
                n += 1

            return val_loss / n
        else:
            with torch.inference_mode():
                val_loss += loss_fn(val_data, None).item()
            return val_loss

    def finalise(self):
        """Method to finalise the flow before using it for inference."""
        logger.debug("Finalising model before inference.")
        self.model.finalise()

    def train(
        self,
        samples,
        weights=None,
        conditional=None,
        max_epochs=None,
        patience=None,
        output=None,
        val_size=None,
        plot=True,
    ):
        """
        Train the flow on a set of samples.

        Allows for training parameters to specified instead of those
        given in initial config.

        Parameters
        ----------
        samples : ndarray
            Unstructured numpy array containing data to train on
        weights : array_like
            Array of weights to use with the weight KL when computing the loss.
        max_epochs : int, optional
            Maximum number of epochs that is used instead of value
            in the configuration.
        patience : int, optional
            Patience in number of epochs that is used instead of value
            in the configuration.
        val_size : float, optional
           Fraction of the samples to use for validation
        output : str, optional
            Path to output directory that is used instead of the path
            specified when the object is initialised
        plot : bool, optional
            Boolean to enable or disable plotting the loss function

        Returns
        -------
        history : dict
            Dictionary that contains the training and validation losses.
        """
        if not self.initialised:
            logger.debug("Initialising")
            self.initialise()

        if not np.isfinite(samples).all():
            raise ValueError("Training data is not finite")

        logger.debug("Data summary:")
        logger.debug(f"Mean: {np.mean(samples, axis=0)}")
        logger.debug(f"Standard deviation: {np.std(samples, axis=0, ddof=1)}")

        if output is None:
            output = self.output
        else:
            os.makedirs(output, exist_ok=True)

        if val_size is None:
            val_size = self.training_config["val_size"]

        if val_size == 0.0:
            validate = False
        else:
            validate = True

        if self.noise_type == "adaptive":
            noise_scale = self.noise_scale * np.mean(
                compute_minimum_distances(samples)
            )
            logger.debug(f"Using adaptive scale: {noise_scale:.3f}")
        elif self.noise_type == "constant":
            noise_scale = self.noise_scale
        else:
            logger.debug("No noise will be added in training")
            noise_scale = None

        self.move_to(self.device)

        weighted = True if weights is not None else False
        is_conditional = True if conditional is not None else False
        use_dataloader = (
            self.training_config["use_dataloader"]
            or weighted
            or is_conditional
        )

        train_data, val_data, _ = self.prep_data(
            samples,
            val_size=val_size,
            batch_size=self.training_config["batch_size"],
            weights=weights,
            conditional=conditional,
            use_dataloader=use_dataloader,
        )

        if max_epochs is None:
            max_epochs = self.training_config["max_epochs"]
        if self.training_config["annealing"]:
            self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                self._optimiser, max_epochs
            )
        if patience is None:
            patience = self.training_config["patience"]
        best_epoch = 0
        best_val_loss = np.inf
        best_model = copy.deepcopy(self.model.state_dict())
        logger.debug("Starting training")
        logger.debug("Training parameters:")
        logger.debug(f"Max. epochs: {max_epochs}")
        logger.debug(f"Patience: {patience}")
        logger.debug(f"Training with {samples.shape[0]} samples")

        history = dict(loss=[], val_loss=[])

        current_weights_file = os.path.join(output, "model.pt")

        for epoch in range(1, max_epochs + 1):
            loss = self._train(
                train_data,
                noise_scale=noise_scale,
                is_dataloader=use_dataloader,
                weighted=weighted,
                is_conditional=is_conditional,
            )
            val_loss = self._validate(
                val_data,
                is_dataloader=use_dataloader,
                weighted=weighted,
                is_conditional=is_conditional,
            )
            history["loss"].append(loss)
            history["val_loss"].append(val_loss)

            if validate and (val_loss < best_val_loss):
                best_epoch = epoch
                best_val_loss = val_loss
                best_model = copy.deepcopy(self.model.state_dict())

            if not epoch % 50:
                logger.debug(
                    f"Epoch {epoch}: loss: {loss:.3} val loss: {val_loss:.3}"
                )

            if validate and (epoch - best_epoch > patience):
                logger.debug(f"Epoch {epoch}: Reached patience")
                break

        logger.debug("Finished training")

        # Make sure cache for LU is reset.
        self.model.train()
        self.model.eval()
        if validate:
            logger.debug(f"Loading best model from epoch {best_epoch}")
            self.model.load_state_dict(best_model)
        self.finalise()

        self.save_weights(current_weights_file)
        self.move_to(self.inference_device)
        self.model.eval()

        if plot:
            plot_loss(
                epoch, history, filename=os.path.join(output, "loss.png")
            )
        return history

    def save_weights(self, weights_file):
        """
        Save the weights file. If the file already exists move it to
        ``<weights_file>.old`` and then save the file.

        Parameters
        ----------
        weights_file : str
            Path to to file to save weights. Recommended file type is ``.pt``.
        """
        if os.path.exists(weights_file):
            shutil.move(weights_file, weights_file + ".old")

        torch.save(self.model.state_dict(), weights_file)
        self.weights_file = weights_file

    def load_weights(self, weights_file):
        """
        Load weights for the model and initialises the model if it is not
        initialised. The weights_file attribute is also updated.

        Model is loaded in evaluation mode (``model.eval()``)

        Parameters
        ----------
        weights_files : str
            Path to weights file
        """
        # TODO: these two methods are basically the same
        if not self.initialised:
            self.initialise()
        self.model.load_state_dict(torch.load(weights_file, weights_only=True))
        self.model.eval()
        self.weights_file = weights_file

    def reload_weights(self, weights_file):
        """
        Tries to the load the weights file and if not, tries to load
        the weights file stored internally.

        Parameters
        ----------
        weights_files : str
            Path to weights file
        """
        if weights_file is None:
            weights_file = self.weights_file
        logger.debug(f"Reloading weights from {weights_file}")
        self.load_weights(weights_file)

    def reset_model(self, weights=True, permutations=False):
        """
        Reset the weights of the model and optimiser.

        Parameters
        ----------
        weights : bool, optional
            If true the model weights are reset.
        permutations : bool, optional
            If true any permutations (linear transforms) are reset.
        """
        if not any([weights, permutations]):
            logger.debug("Nothing to reset")
            return
        if weights and permutations:
            logger.debug("Complete reset of model")
            self.model = configure_model(self.flow_config)
        elif weights:
            self.model.apply(reset_weights)
            logger.debug("Reset weights")
        elif permutations:
            self.model.apply(reset_permutations)
            logger.debug("Reset linear transforms")
        self._optimiser = self.get_optimiser()
        logger.debug("Resetting optimiser")

    def numpy_array_to_tensor(self, array: np.ndarray, /) -> torch.Tensor:
        """Convert a numpy array to a tensor and move it to the device"""
        return (
            torch.from_numpy(array)
            .type(torch.get_default_dtype())
            .to(self.model.device)
        )

    def forward_and_log_prob(
        self, x: np.ndarray, conditional: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Forward pass through the model and return the samples in the latent
        space with their log probabilities

        Parameters
        ----------
        x : ndarray
            Array of samples

        Returns
        -------
        z : ndarray
            Samples in the latent space
        log_prob : ndarray
            Log probabilities for each samples
        """
        x = self.numpy_array_to_tensor(x)
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        self.model.eval()
        with torch.inference_mode():
            z, log_prob = self.model.forward_and_log_prob(
                x, context=conditional
            )
        z = z.detach().cpu().numpy().astype(np.float64)
        log_prob = log_prob.detach().cpu().numpy().astype(np.float64)
        return z, log_prob

    def inverse(
        self, z: np.ndarray, conditional: Optional[np.ndarray] = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Apply the inverse transform.

        Parameters
        ----------
        z :
            Array of samples in the latent space.
        conditional :
            Conditional inputs.

        Returns
        -------
        x :
            Array of samples in the X-prime space.
        log_j :
            Log-Jacobian determinant for each sample.
        """
        z = self.numpy_array_to_tensor(z)
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        self.model.eval()
        with torch.inference_mode():
            x, log_j = self.model.inverse(z, context=conditional)
        x = x.detach().cpu().numpy().astype(np.float64)
        log_j = log_j.detach().cpu().numpy().astype(np.float64)
        return x, log_j

    def log_prob(
        self, x: np.ndarray, conditional: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Compute the log-probability of a sample.

        Parameters
        ----------
        x : ndarray
            Array of samples in the X-prime space.

        Returns
        -------
        ndarray
            Array of log-probabilities.
        """
        x = self.numpy_array_to_tensor(x)
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        self.model.eval()
        with torch.inference_mode():
            log_prob = self.model.log_prob(x, context=conditional)
        log_prob = log_prob.cpu().numpy().astype(np.float64)
        return log_prob

    def sample(
        self, n: int = 1, conditional: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Sample from the flow.

        Parameters
        ----------
        n : int
            Number of samples to draw
        conditional : numpy.ndarray
            Array of conditional inputs

        Returns
        -------
        numpy.ndarray
            Array of samples
        """
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        with torch.inference_mode():
            x = self.model.sample(int(n), context=conditional)
        return x.cpu().numpy().astype(np.float64)

    def sample_latent_distribution(self, n: int = 1) -> np.ndarray:
        """Sample from the latent distribution.

        Parameters
        ----------
        n : int
            Number of samples to draw

        Returns
        -------
        numpy.ndarray
            Array of samples
        """
        with torch.inference_mode():
            z = self.model.sample_latent_distribution(n)
        return z.cpu().numpy().astype(np.float64)

    def sample_and_log_prob(
        self, N=1, z=None, alt_dist=None, conditional=None
    ):
        """
        Generate samples from samples drawn from the base distribution or
        and alternative distribution from provided latent samples

        Parameters
        ----------
        N : int, optional
            Number of samples to draw if z is not specified
        z : ndarray, optional
            Array of latent samples to map the the data space, if ``alt_dist``
            is not specified they are assumed to be drawn from the base
            distribution of the flow.
        alt_dist : :obj:`glasflow.nflows.distribution.Distribution`
            Distribution object from which the latent samples z were
            drawn from. Must have a ``log_prob`` method that accepts an
            instance of ``torch.Tensor``

        Returns
        -------
        samples : ndarray
            Array containing samples in the latent space.
        log_prob : ndarray
            Array containing the log probability that corresponds to each
            sample.
        """
        if self.model is None:
            raise RuntimeError("Model is not initialised yet!")
        if self.model.training:
            self.model.eval()
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        if z is None:
            with torch.inference_mode():
                x, log_prob = self.model.sample_and_log_prob(
                    int(N), context=conditional
                )
        else:
            if alt_dist is not None:
                log_prob_fn = alt_dist.log_prob
            else:
                log_prob_fn = self.model.base_distribution_log_prob

            with torch.inference_mode():
                if isinstance(z, np.ndarray):
                    z = self.numpy_array_to_tensor(z)
                log_prob = log_prob_fn(z)
                x, log_J = self.model.inverse(z, context=conditional)
                log_prob -= log_J

        x = x.detach().cpu().numpy().astype(np.float64)
        log_prob = log_prob.detach().cpu().numpy().astype(np.float64)
        return x, log_prob

    def __getstate__(self):
        state = self.__dict__.copy()
        state["initialised"] = False
        del state["_optimiser"]
        del state["model"]
        # This may have classes etc
        del state["flow_config"]
        return state



from unittest.mock import create_autospec

import numpy as np
import pytest

from nessai.reparameterisations.rescale import PrePostRescalingMixin
from nessai.utils.rescaling import rescaling_functions


@pytest.fixture
def reparam():
    return create_autospec(PrePostRescalingMixin)


def test_default_pre_rescaling(reparam):
    """Assert the default pre-rescaling is the identity"""
    x = np.array([1, 2, 3])
    expected_log_j = np.zeros(3)
    x_out, log_j = PrePostRescalingMixin.pre_rescaling(reparam, x)
    x_out_inv, log_j_inv = PrePostRescalingMixin.pre_rescaling_inv(reparam, x)

    np.testing.assert_array_equal(x_out, x)
    np.testing.assert_array_equal(x_out_inv, x)
    np.testing.assert_array_equal(log_j, expected_log_j)
    np.testing.assert_array_equal(log_j_inv, expected_log_j)


def test_default_post_rescaling(reparam):
    """Assert the default post-rescaling is the identity"""
    x = np.array([1, 2, 3])
    expected_log_j = np.zeros(3)
    x_out, log_j = PrePostRescalingMixin.post_rescaling(reparam, x)
    x_out_inv, log_j_inv = PrePostRescalingMixin.post_rescaling_inv(reparam, x)

    np.testing.assert_array_equal(x_out, x)
    np.testing.assert_array_equal(x_out_inv, x)
    np.testing.assert_array_equal(log_j, expected_log_j)
    np.testing.assert_array_equal(log_j_inv, expected_log_j)


def test_configure_pre_rescaling_none(reparam):
    """Test the configuration of the pre-rescaling if it is None"""
    PrePostRescalingMixin.configure_pre_rescaling(reparam, None)
    assert reparam.has_pre_rescaling is False


def test_configure_post_rescaling_none(reparam):
    """Test the configuration of the post-rescaling if it is None"""
    PrePostRescalingMixin.configure_post_rescaling(reparam, None)
    assert reparam.has_post_rescaling is False


def test_pre_rescaling_with_functions(reparam):
    """Assert that specifying functions works as intended"""
    rescaling = (np.exp, np.log)
    PrePostRescalingMixin.configure_pre_rescaling(reparam, rescaling)
    assert reparam.has_pre_rescaling is True
    assert reparam.pre_rescaling is np.exp
    assert reparam.pre_rescaling_inv is np.log


def test_post_rescaling_with_functions(reparam):
    """Assert that specifying functions works as intended"""
    rescaling = (np.exp, np.log)
    PrePostRescalingMixin.configure_post_rescaling(reparam, rescaling)
    assert reparam.has_post_rescaling is True
    assert reparam.has_prime_prior is False
    assert reparam.post_rescaling is np.exp
    assert reparam.post_rescaling_inv is np.log


@pytest.mark.parametrize("rescaling", ["logit", "inv_gaussian_cdf"])
def test_pre_rescaling_with_str(reparam, rescaling):
    """Assert that specifying a str works as intended"""
    PrePostRescalingMixin.configure_pre_rescaling(reparam, rescaling)
    assert reparam.has_pre_rescaling is True
    assert reparam.pre_rescaling is rescaling_functions[rescaling][0]
    assert reparam.pre_rescaling_inv is rescaling_functions[rescaling][1]


@pytest.mark.parametrize("rescaling", ["log", "logit"])
def test_post_rescaling_with_str(reparam, rescaling):
    """Assert that specifying a str works as intended"""
    reparam._update = False
    reparam.parameters = ["x"]

    PrePostRescalingMixin.configure_post_rescaling(reparam, rescaling)
    assert reparam.has_post_rescaling is True
    assert reparam.has_prime_prior is False
    assert reparam.post_rescaling is rescaling_functions[rescaling][0]
    assert reparam.post_rescaling_inv is rescaling_functions[rescaling][1]


def test_pre_rescaling_with_invalid_str(reparam):
    """Assert an error is raised if the rescaling is not recognised"""
    rescaling = "not_a_rescaling"
    with pytest.raises(RuntimeError) as excinfo:
        PrePostRescalingMixin.configure_pre_rescaling(reparam, rescaling)
    assert "Unknown rescaling function: not_a_rescaling" in str(excinfo.value)


def test_post_rescaling_with_invalid_str(reparam):
    """Assert an error is raised if the rescaling is not recognised"""
    rescaling = "not_a_rescaling"
    with pytest.raises(RuntimeError) as excinfo:
        PrePostRescalingMixin.configure_post_rescaling(reparam, rescaling)
    assert "Unknown rescaling function: not_a_rescaling" in str(excinfo.value)


def test_pre_rescaling_invalid_input(reparam):
    """Assert an error is raised if the input isn't a str or tuple"""
    with pytest.raises(RuntimeError) as excinfo:
        PrePostRescalingMixin.configure_pre_rescaling(reparam, (np.exp,))
    assert "Pre-rescaling must be a str or tuple" in str(excinfo.value)


def test_post_rescaling_invalid_input(reparam):
    """Assert an error is raised if the input isn't a str or tuple"""
    with pytest.raises(RuntimeError) as excinfo:
        PrePostRescalingMixin.configure_post_rescaling(reparam, (np.exp,))
    assert "Post-rescaling must be a str or tuple" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Base object for all proposal classes.
"""

import datetime
import logging
import os
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np

logger = logging.getLogger(__name__)


class Proposal(ABC):
    """
    Base proposal object

    Parameters
    ----------
    model: obj
        User-defined model
    rng: np.random.Generator, optional
        Random number generator. If not provided, a new generator is created.
    """

    def __init__(self, model, rng: Optional[np.random.Generator] = None):
        self.model = model
        if rng is None:
            logger.debug("No rng specified, using the default rng.")
            rng = np.random.default_rng()
        self.rng = rng
        self.populated = True
        self._initialised = False
        self.training_count = 0
        self.population_acceptance = None
        self.population_time = datetime.timedelta()
        self.r = np.nan
        self.samples = []
        self.indices = []
        self._checked_population = True

    @property
    def initialised(self):
        """Boolean that indicates if the proposal is initialised or not."""
        return self._initialised

    @initialised.setter
    def initialised(self, boolean):
        """Setter for initialised"""
        if boolean:
            self._initialised = boolean
        else:
            self._initialised = boolean

    def initialise(self):
        """
        Initialise the proposal
        """
        self.initialised = True

    def update_output(self, output: str) -> None:
        """
        Update the output directory.

        Only updates the output if the proposal has an output attribute.

        Parameters
        ----------
        output: str
            Path to the output directory
        """
        if hasattr(self, "output"):
            logger.debug(f"Updating output directory to {output}")
            self.output = output
            os.makedirs(self.output, exist_ok=True)
        else:
            logger.debug("No output directory to update")

    def evaluate_likelihoods(self):
        """Evaluate the likelihoods for the pool of live points."""
        self.samples["logL"] = self.model.batch_evaluate_log_likelihood(
            self.samples
        )

    @abstractmethod
    def draw(self, old_param):
        """
        New a new point given the old point
        """
        raise NotImplementedError

    def train(self, x, **kwargs):
        """
        Train the proposal method

        Parameters
        ----------
        x: array_like
            Array of live points to use for training
        kwargs:
            Any of keyword arguments
        """
        logger.error("This proposal method cannot be trained")

    def resume(self, model):
        """
        Resume the proposal with the model
        """
        self.model = model

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["model"]
        return state



# -*- coding: utf-8 -*-
from unittest.mock import create_autospec

import pytest

from nessai.proposal.flowproposal.base import BaseFlowProposal


@pytest.fixture(params=[True, False])
def map_to_unit_hypercube(request):
    return request.param


@pytest.fixture()
def proposal(rng):
    proposal = create_autospec(BaseFlowProposal, rng=rng)
    proposal._initialised = False
    proposal.map_to_unit_hypercube = False
    return proposal



# -*- coding: utf-8 -*-
"""
Utilities for handling the reparameterisations.
"""

import copy
import logging
from dataclasses import dataclass, field
from typing import Any

from ..utils.entry_points import get_entry_points
from .base import Reparameterisation
from .null import NullReparameterisation

logger = logging.getLogger(__name__)


class ReparameterisationError(RuntimeError):
    """Exception for reparameterisation errors"""


@dataclass(frozen=True)
class KnownReparameterisation:
    """Dataclass to store the reparameterisation class and keyword arguments"""

    name: str
    class_fn: Reparameterisation
    keyword_arguments: dict[str:Any] = field(default_factory=dict)


class ReparameterisationDict(dict):
    """Dictionary of reparameterisations

    This dictionary is used to store the known reparameterisations and
    provides a method to add new reparameterisations.
    """

    def add_reparameterisation(self, name, class_fn, keyword_arguments=None):
        """Add a new reparameterisation to the dictionary

        Parameters
        ----------
        name : str
            Name of the reparameterisation.
        class_fn : Reparameterisation
            Reparameterisation class.
        keyword_arguments : dict, optional
            Keyword arguments for the reparameterisation.
        """
        if keyword_arguments is None:
            keyword_arguments = {}
        if name in self:
            raise ValueError(f"Reparameterisation {name} already exists")
        self[name] = KnownReparameterisation(name, class_fn, keyword_arguments)

    def add_external_reparameterisations(self, group):
        entry_points = get_entry_points(group)
        for ep in entry_points.values():
            reparam = ep.load()
            if not isinstance(reparam, KnownReparameterisation):
                raise RuntimeError(
                    f"Invalid external reparameterisation: {reparam}"
                )
            elif reparam.name in self:
                raise ValueError(
                    f"Reparameterisation {reparam.name} already exists"
                )
            logger.debug(f"Adding external reparameterisation: {reparam}")
            self[reparam.name] = reparam


def get_reparameterisation(reparameterisation, defaults=None):
    """Function to get a reparameterisation class from a name

    Parameters
    ----------
    reparameterisation : str, \
            :obj:`nessai.reparameterisations.Reparameterisation`
        Name of the reparameterisations to return or a class that inherits from
        :obj:`~nessai.reparameterisations.Reparameterisation`
    defaults : dict, optional
        Dictionary of known reparameterisations that overrides the defaults.

    Returns
    -------
    :obj:`nessai.reparameteristaions.Reparameterisation`
        Reparameterisation class.
    dict
        Keyword arguments for the specific reparameterisation.
    """
    if defaults is None:
        from . import default_reparameterisations

        defaults = default_reparameterisations

    if isinstance(reparameterisation, str):
        reparam = defaults.get(reparameterisation, None)
        if reparam is None:
            raise ValueError(
                f"Unknown reparameterisation: {reparameterisation}. "
                f"Known reparameterisations are: {list(defaults.keys())}."
            )
        else:
            return reparam.class_fn, copy.deepcopy(reparam.keyword_arguments)
    elif reparameterisation is None:
        return NullReparameterisation, {}
    elif isinstance(reparameterisation, type) and issubclass(
        reparameterisation, Reparameterisation
    ):
        return reparameterisation, {}
    else:
        raise TypeError(
            "Reparameterisation must be a str, None, or class that "
            "inherits from `Reparameterisation`"
        )



import datetime
import logging
import os

import matplotlib.pyplot as plt
import numpy as np

from ....livepoint import (
    live_points_to_array,
)
from ....proposal.flowproposal.base import BaseFlowProposal
from .steps import (
    KNOWN_STEPS,
)

logger = logging.getLogger(__name__)


class MCMCFlowProposal(BaseFlowProposal):
    """Version of FlowProposal that uses MCMC instead of rejection sampling"""

    def __init__(
        self,
        model,
        n_steps=10,
        n_accept=None,
        step_type: str = "gaussian",
        step_kwargs: dict = None,
        plot_chain: bool = False,
        plot_history: bool = False,
        enforce_likelihood_threshold: bool = True,
        ensemble_fraction: float = 0.5,
        **kwargs,
    ):
        self.n_steps = n_steps
        self.n_accept = n_accept
        self.step_type = step_type
        self._plot_chain = plot_chain
        self._plot_history = plot_history
        self.enforce_likelihood_threshold = enforce_likelihood_threshold
        self.ensemble_fraction = ensemble_fraction
        self.mcmc_history = {
            "acceptance": [],
            "n_steps": [],
        }
        self.step_kwargs = step_kwargs or {}
        super().__init__(model, **kwargs)

    def initialise(self, resumed: bool = False):
        """Initialise the proposal.

        This includes setting up the MCMC step.
        """
        super().initialise(resumed=resumed)
        StepClass = KNOWN_STEPS.get(self.step_type)
        logger.debug(
            f"Using step type: {StepClass} with kwargs: {self.step_kwargs}"
        )
        self.step = StepClass(
            dims=self.rescaled_dims, rng=self.rng, **self.step_kwargs
        )

    def plot_chain(self, chains):
        nsteps, nchains, ndims = chains.shape

        fig, axs = plt.subplots(ndims, 1, figsize=(4, 10))
        for i in range(nchains):
            for j in range(ndims):
                axs[j].plot(chains[:, i, j])
        fig.savefig(
            os.path.join(self.output, f"chain_{self.populated_count}.png")
        )
        plt.close(fig)

    def plot_history(self):
        """Plot the history of MCMC acceptance and number of steps.

        This is useful for diagnosing the performance of the MCMC proposal over
        the course of the run.
        """
        fig, axs = plt.subplots(2, 1, sharex=True)
        axs[0].plot(self.mcmc_history["acceptance"])
        axs[0].set_ylabel("Acceptance")
        axs[1].plot(self.mcmc_history["n_steps"])
        axs[1].set_ylabel("Number of steps")
        axs[-1].set_xlabel("Iteration")
        plt.tight_layout()
        fig.savefig(os.path.join(self.output, "mcmc_history.png"))
        plt.close(fig)

    def x_prime_log_prior(self, x):
        raise RuntimeError(
            "MCMCFlowProposal does not support using x-prime priors"
        )

    def populate(
        self,
        worst_point,
        n_samples=1000,
        plot=True,
    ):
        st = datetime.datetime.now()
        if not self.initialised:
            raise RuntimeError(
                "Proposal has not been initialised. "
                "Try calling `initialise()` first."
            )

        log_l_threshold = worst_point["logL"].copy()

        # Ensemble points
        x_prime_array = live_points_to_array(
            self.training_data_prime,
            self.prime_parameters,
            copy=True,
        )
        self.rng.shuffle(x_prime_array)
        z_ensemble, _ = self.flow.forward_and_log_prob(x_prime_array)

        ensemble_size = int(self.ensemble_fraction * self.training_data.size)
        self.step.update_ensemble(z_ensemble[:ensemble_size])

        poolsize = n_samples if n_samples is not None else self.poolsize
        n_walkers = min(poolsize, self.training_data.size - ensemble_size)

        # Initial points
        z_current = z_ensemble[ensemble_size : (ensemble_size + n_walkers)]
        x_current, log_j_current = self.backward_pass(
            z_current, return_unit_hypercube=self.map_to_unit_hypercube
        )
        if self.map_to_unit_hypercube:
            log_p_current = self.unit_hypercube_log_prior(x_current)
        else:
            log_p_current = self.log_prior(x_current)
        x_current["logL"] = self.model.batch_evaluate_log_likelihood(
            x_current, unit_hypercube=self.map_to_unit_hypercube
        )

        z_chain = np.empty((self.n_steps, n_walkers, z_current.shape[-1]))
        z_chain[0] = z_current

        z_new_history = []
        n_accept = np.zeros(n_walkers)
        n_reject = np.zeros(n_walkers)

        for i in range(self.n_steps):
            z_new, log_j_step = self.step(z_current)
            z_new_history.append(z_new)

            x_new, log_j_flow = self.backward_pass(
                z_new,
                rescale=True,
                return_unit_hypercube=self.map_to_unit_hypercube,
            )
            if self.map_to_unit_hypercube:
                log_p = self.unit_hypercube_log_prior(x_new)
            else:
                log_p = self.log_prior(x_new)
            finite_prior = np.isfinite(log_p)

            # Jacobian should include flow and step
            log_j_new = log_j_step + log_j_flow
            # Calculate acceptance
            log_factor = log_p + log_j_new - log_p_current - log_j_current
            log_u = np.log(self.rng.random(n_walkers))
            accept = (log_factor > log_u) & finite_prior
            # Only evaluate function where log-prior is finite
            # Default is NaN, so will not pass threshold.
            if self.enforce_likelihood_threshold:
                x_new["logL"][finite_prior] = (
                    self.model.batch_evaluate_log_likelihood(
                        x_new[finite_prior],
                        unit_hypercube=self.map_to_unit_hypercube,
                    )
                )
                logl_accept = x_new["logL"] > log_l_threshold
                accept &= logl_accept

            x_current[accept] = x_new[accept]
            z_current[accept] = z_new[accept]
            # Only include the log-jacobian for the flow
            log_j_current[accept] = log_j_flow[accept]
            n_accept += accept
            n_reject += 1 - accept
            z_chain[i] = z_current
            if self.n_accept is not None and n_accept.mean() > self.n_accept:
                n_steps = i
                break
        else:
            n_steps = self.n_steps
            if self.n_accept is not None:
                logger.warning(
                    (
                        f"Reached max steps ({self.n_steps}) with "
                        f"n_accept={self.n_accept}!"
                    )
                )

        keep = n_accept > 0
        logger.debug(f"Replacing {n_walkers - keep.sum()} walkers")
        x_current = x_current[keep]

        x_current["logL"] = self.model.batch_evaluate_log_likelihood(
            x_current, unit_hypercube=self.map_to_unit_hypercube
        )

        z_new_history = np.array(z_new_history)
        z_chain = z_chain[:n_steps]
        self.step.update_stats(
            n_accept=n_accept.mean(),
            n_reject=n_reject.mean(),
        )

        self.population_time += datetime.datetime.now() - st
        if len(x_current) == 0:
            logger.warning("No samples accepted!")
            return

        self.samples = self.convert_to_samples(x_current)
        if self._plot_chain and plot:
            self.plot_chain(z_chain)
        if self._plot_pool and plot:
            self.plot_pool(self.samples)

        acceptance = n_accept.mean() / (n_accept.mean() + n_reject.mean())
        self.mcmc_history["acceptance"].append(acceptance)
        self.mcmc_history["n_steps"].append(n_steps)
        if self._plot_history and plot:
            self.plot_history()

        self.population_acceptance = self.mcmc_history["acceptance"][-1]

        logger.debug(f"MCMC acceptance: {self.population_acceptance}")
        self.indices = self.rng.permutation(self.samples.size).tolist()
        self.populated_count += 1
        self.populated = True
        self._checked_population = False



#!/usr/bin/env python
"""
Script to generate example plots for a run with nessai
"""

from nessai_models import Gaussian

from nessai.flowsampler import FlowSampler

output_dir = "example_run"
model = Gaussian()

fs = FlowSampler(
    model, nlive=500, output=output_dir, checkpointing=False, resume=False
)
fs.run()
